from importlib import resources
import logging
import numpy as np
import matplotlib
from matplotlib.colors import Normalize
from mpl_toolkits.mplot3d.axes3d import Axes3D
import warnings

import time

from PyQt6 import QtWidgets, uic, QtCore
from PyQt6.QtCore import QObject, QTimer, QThread, pyqtSignal, QCoreApplication
import multiprocessing as mp

from bmlab.session import Session
from bmlab.fits import lorentz

from bmicro.gui.mpl import MplCanvas

from bmlab.controllers import EvaluationController

logger = logging.getLogger(__name__)


class Worker(QObject):
    finished = pyqtSignal()

    def __init__(self, fkw):
        super().__init__()
        self.evaluation_controller = EvaluationController()
        self.fkw = fkw

    def run(self):
        self.evaluation_controller.evaluate(**self.fkw)
        self.finished.emit()


class EvaluationView(QtWidgets.QWidget):
    """
    Class for the evaluation widget
    """

    def __init__(self, *args, **kwargs):
        super(EvaluationView, self).__init__(*args, **kwargs)

        ref = resources.files('bmicro.gui.evaluation') / 'evaluation_view.ui'
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self)

        self.mplcanvas = MplCanvas(self.image_widget,
                                   toolbar=('Home', 'Pan', 'Zoom'))
        self.mplcanvas.get_figure().canvas.mpl_connect(
            'button_press_event', self.on_click_image)
        self.plot = self.mplcanvas.get_figure().add_subplot(111)
        self.image_map = None
        self.colorbar = None

        self.image_spectrum_dialog = None
        self.isd_image_canvas = None
        self.isd_image_plot = None
        self.isd_image_map = None
        self.isd_image_colorbar = None
        self.isd_spectrum_canvas = None
        self.isd_spectrum_plot = None

        self.button_evaluate.released.connect(self.evaluate)

        self.setup_parameter_selection_combobox()

        self.combobox_parameter.currentIndexChanged.connect(
            self.on_select_parameter)
        self.combobox_peak_number.currentIndexChanged.connect(
            self.on_select_parameter)

        self.aspect_ratio.clicked.connect(
            self.refresh_plot)

        self.autoscale.clicked.connect(
            self.on_scale_changed)
        self.value_min.valueChanged.connect(
            self.on_scale_changed)
        self.value_max.valueChanged.connect(
            self.on_scale_changed)

        self.evaluation_controller = EvaluationController()

        self.evaluation_abort = mp.Value('I', False, lock=True)
        self.evaluation_running = False

        self.session = Session.get_instance()

        self.evaluation_timer = QTimer()
        self.evaluation_timer.timeout.connect(self.refresh_ui)
        self.count = None
        self.max_count = None
        self.thread = None
        self.worker = None
        # Currently used to determine if we should update the plot
        # Might not be necessary anymore once the plot is fast enough.
        self.plot_count = 0

        self.bounds_table.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.bounds_table.verticalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.bounds_table.cellChanged.connect(self.boundsChanged)

        self.nrBrillouinPeaks_1.toggled.connect(
            lambda: self.setNrBrillouinPeaks(1))
        self.nrBrillouinPeaks_2.toggled.connect(
            lambda: self.setNrBrillouinPeaks(2))
        self.nrBrillouinPeaks_4.toggled.connect(
            lambda: self.setNrBrillouinPeaks(4))

    def update_ui(self):
        session = Session.get_instance()
        evm = session.evaluation_model()
        if evm is None:
            return

        if evm.nr_brillouin_peaks == 1:
            self.nrBrillouinPeaks_1.setChecked(True)
            self.bounds_table.setEnabled(False)
        elif evm.nr_brillouin_peaks == 2:
            self.nrBrillouinPeaks_2.setChecked(True)
            self.bounds_table.setEnabled(True)
        elif evm.nr_brillouin_peaks == 4:
            self.nrBrillouinPeaks_4.setChecked(True)
            self.bounds_table.setEnabled(True)

        self.updateBoundsTable()
        self.setup_parameter_selection_combobox()
        self.refresh_plot()

    def updateBoundsTable(self):
        session = Session.get_instance()
        evm = session.evaluation_model()
        if evm is None:
            self.bounds_table.setRowCount(0)
            self.bounds_table.setEnabled(False)
            return
        bounds_w0 = evm.bounds_w0
        bounds_fwhm = evm.bounds_fwhm
        if bounds_w0 is None or bounds_fwhm is None:
            self.bounds_table.setRowCount(0)
        else:
            self.bounds_table.setColumnCount(4)
            self.bounds_table.setRowCount(len(bounds_w0))
            for i, bound in enumerate(bounds_w0):
                item = QtWidgets.QTableWidgetItem(str(bound[0]))
                self.bounds_table.setItem(i, 0, item)
                item = QtWidgets.QTableWidgetItem(str(bound[1]))
                self.bounds_table.setItem(i, 1, item)
            for i, bound in enumerate(bounds_fwhm):
                item = QtWidgets.QTableWidgetItem(str(bound[0]))
                self.bounds_table.setItem(i, 2, item)
                item = QtWidgets.QTableWidgetItem(str(bound[1]))
                self.bounds_table.setItem(i, 3, item)

        self.bounds_table.setEnabled(evm.nr_brillouin_peaks > 1)

    def on_click_image(self, event):
        """
        Action triggered when user clicks Brillouin map.

        Parameters
        ----------
        event: matplotlib event object
            The mouse click event.
        """
        # If the click is outside the axes, skip it
        if event.inaxes is None:
            return
        # If we don't have a session we have not loaded data yet
        session = Session.get_instance()
        if session is None:
            return

        # Get the resolution of the measurement
        resolution = self.session.get_payload_resolution()
        if resolution is None:
            return

        # Get the positions and normalize them
        positions = list(self.session.get_payload_positions().values())
        for position in positions:
            position -= np.nanmean(position)

        # Get the indices of the dimensions to check
        idx = [idx for idx, dim in enumerate(resolution) if dim > 1]

        # Only works for 1D and 2D plots
        if len(idx) < 1 or len(idx) > 2:
            return

        # Determine the indices of the click in the positions arrays
        click_pos = (event.xdata, event.ydata)
        indices = np.zeros(len(resolution), dtype="int")
        for ind, p_ind in enumerate(idx):
            dslice = [slice(None) if p_ind == i else 0
                      for i, res in enumerate(resolution)]
            r = abs(positions[p_ind][tuple(dslice)] - click_pos[ind])
            indices[p_ind] = int(np.argmin(r))

        # Convert indices to key
        image_key = self.evaluation_controller\
            .get_key_from_indices(resolution, *indices)

        # If the modal is not open yet, open it
        self.open_image_spectrum()

        # Show position and key
        self.image_spectrum_dialog.image_spectrum_label.setText(
            "Position "
            f"x: {positions[0][tuple(indices)]:#0.1f} [µm], "
            f"y: {positions[1][tuple(indices)]:#0.1f} [µm], "
            f"z: {positions[2][tuple(indices)]:#0.1f} [µm], "
            f"image key: {image_key}"
        )

        # Get the image plot it
        image = session.get_payload_image(image_key, 0)
        if image is not None:
            if isinstance(self.isd_image_map, matplotlib.image.AxesImage):
                self.isd_image_map.set_data(image.T)
            else:
                self.isd_image_map = self.isd_image_plot.imshow(
                    image.T, origin='lower', vmin=100, vmax=300
                )
                self.isd_image_colorbar = \
                    self.isd_image_canvas\
                        .get_figure().colorbar(self.isd_image_map)
        self.isd_image_canvas.draw()

        # Get spectrum and plot it
        self.isd_spectrum_plot.cla()
        spectra = session.evaluation_model().get_spectra(image_key)
        cm = session.calibration_model()
        payload_time = session.get_payload_time(image_key)
        frequencies = cm.get_frequencies_by_time(payload_time)

        if spectra is not None and frequencies is not None:
            image_nr = 0
            spectrum = np.nanmean(spectra, image_nr)

            # Also try to get the fit
            brillouin_fits, rayleigh_fits =\
                self.evaluation_controller.get_fits(image_key)

            # Show the measured data
            self.isd_spectrum_plot.plot(1e-9 * frequencies[0],
                                        spectrum, color='tab:blue')
            self.isd_spectrum_plot.set_xlabel('$f$ [GHz]')

            pm = session.peak_selection_model()
            evm = session.evaluation_model()
            if pm is not None and evm is not None:
                # Show the Brillouin peaks
                brillouin_regions = pm.get_brillouin_regions()
                # Iterate over the regions
                for region_nr in range(brillouin_fits[0].shape[1]):
                    x = np.linspace(
                        brillouin_regions[region_nr][0],
                        brillouin_regions[region_nr][1],
                        200
                    )
                    # First entry is always a single-peak fit,
                    # the following entries belong to a multi-peak fit
                    for peak_nr in range(brillouin_fits[0].shape[2]):
                        idx = (image_nr, region_nr, peak_nr)
                        current_fit = lorentz(
                            x,
                            brillouin_fits[0][idx],
                            brillouin_fits[1][idx],
                            brillouin_fits[2][idx]
                        )
                        if peak_nr < 2:
                            y = current_fit
                        else:
                            y += current_fit
                        if peak_nr == 0:
                            color = 'tab:red'
                        else:
                            color = 'tab:orange'
                        # We plot the fit for the first and last entry
                        if peak_nr == 0 or\
                                peak_nr == (brillouin_fits[0].shape[2] - 1):
                            self.isd_spectrum_plot.plot(
                                1e-9 * x,
                                y + brillouin_fits[3][idx],
                                color=color
                            )

                # Show the Rayleigh peaks
                rayleigh_regions = pm.get_rayleigh_regions()
                # Iterate over the regions
                for region_nr in range(rayleigh_fits[0].shape[1]):
                    x = np.linspace(
                        rayleigh_regions[region_nr][0],
                        rayleigh_regions[region_nr][1],
                        200
                    )
                    idx = (image_nr, region_nr, 0)
                    y = lorentz(
                        x,
                        rayleigh_fits[0][idx],
                        rayleigh_fits[1][idx],
                        rayleigh_fits[2][idx]
                    ) + rayleigh_fits[3][idx]
                    self.isd_spectrum_plot.plot(1e-9 * x,
                                                y, color='tab:purple')

        self.isd_spectrum_canvas.draw()

    def open_image_spectrum(self):
        if self.image_spectrum_dialog is None:
            self.image_spectrum_dialog = QtWidgets.QDialog(
                self,
                QtCore.Qt.WindowType.WindowTitleHint |
                QtCore.Qt.WindowType.WindowCloseButtonHint |
                QtCore.Qt.WindowType.WindowMaximizeButtonHint |
                QtCore.Qt.WindowType.WindowMinimizeButtonHint
            )
            ref = resources.files('bmicro.gui.evaluation') / 'spectrum_view.ui'
            with resources.as_file(ref) as ui_file:
                uic.loadUi(ui_file, self.image_spectrum_dialog)
            self.image_spectrum_dialog\
                .setWindowTitle('Camera image & spectrum')
            self.image_spectrum_dialog.setWindowModality(
                QtCore.Qt.WindowModality.NonModal)

            self.image_spectrum_dialog.show()

            self.isd_image_canvas = MplCanvas(
                self.image_spectrum_dialog.image_widget,
                toolbar=('Home', 'Pan', 'Zoom'))
            self.isd_image_plot =\
                self.isd_image_canvas.get_figure().add_subplot(111)
            self.isd_image_map = None
            self.isd_image_colorbar = None

            self.isd_spectrum_canvas = MplCanvas(
                self.image_spectrum_dialog.spectrum_widget,
                toolbar=('Home', 'Pan', 'Zoom'))
            self.isd_spectrum_plot =\
                self.isd_spectrum_canvas.get_figure().add_subplot(111)

        if self.image_spectrum_dialog.isVisible() is False:
            self.image_spectrum_dialog.setVisible(True)

    def setNrBrillouinPeaks(self, nr_brillouin_peaks):
        if not self.sender().isChecked():
            return
        session = Session.get_instance()
        evm = session.evaluation_model()
        if evm is None:
            return
        evm.setNrBrillouinPeaks(nr_brillouin_peaks)
        self.combobox_peak_number.setEnabled(nr_brillouin_peaks > 1)
        self.combobox_peak_number.blockSignals(True)
        self.combobox_peak_number.clear()
        peak_number_labels = []
        if nr_brillouin_peaks == 1:
            peak_number_labels = ['Single-Peak-Fit']
        elif nr_brillouin_peaks == 2:
            peak_number_labels = [
                'Single-Peak-Fit',
                'Two-Peak-Fit - Peak 1',
                'Two-Peak-Fit - Peak 2',
                'Two-Peak-Fit - Mean',
                'Two-Peak-Fit - Weighted Mean'
            ]
        elif nr_brillouin_peaks == 4:
            peak_number_labels = [
                'Single-Peak-Fit',
                'Four-Peak-Fit - Peak 1',
                'Four-Peak-Fit - Peak 2',
                'Four-Peak-Fit - Peak 3',
                'Four-Peak-Fit - Peak 4',
                'Four-Peak-Fit - Mean',
                'Four-Peak-Fit - Weighted Mean'
            ]
        self.combobox_peak_number.addItems(peak_number_labels)
        self.combobox_peak_number.blockSignals(False)

        self.updateBoundsTable()

    def boundsChanged(self, row, column):
        session = Session.get_instance()
        evm = session.evaluation_model()
        if evm is None:
            return

        if evm.bounds_w0 is not None:
            if column < 2:
                evm.bounds_w0[row][column] =\
                    self.bounds_table.item(row, column).text()
            elif column < 4:
                evm.bounds_fwhm[row][column - 2] =\
                    self.bounds_table.item(row, column).text()

    def clear_plots(self):
        if isinstance(self.colorbar, matplotlib.colorbar.Colorbar):
            self.colorbar.remove()
            self.colorbar = None
        # Clear existing plots
        if isinstance(self.image_map, list):
            for m in self.image_map:
                m.remove()
            self.image_map = None
        if self.image_map is not None:
            self.image_map.remove()
            self.image_map = None

    def reset_ui(self):
        self.evaluation_progress.setValue(0)
        self.clear_plots()
        self.plot.cla()
        self.updateBoundsTable()
        self.nrBrillouinPeaks_1.setChecked(True)
        self.combobox_parameter.clear()
        self.combobox_peak_number.setEnabled(False)
        self.combobox_peak_number.blockSignals(True)
        self.combobox_peak_number.clear()
        self.combobox_peak_number.addItems(['Single-Peak-Fit'])
        self.combobox_peak_number.blockSignals(False)

        self.mplcanvas.draw()

        if self.image_spectrum_dialog is not None\
                and self.image_spectrum_dialog.isVisible():
            self.image_spectrum_dialog.close()

    def setup_parameter_selection_combobox(self):

        session = Session.get_instance()
        evm = session.evaluation_model()
        if evm is None:
            return

        parameters = evm.get_parameter_keys()

        param_labels = []
        for key, parameter in parameters.items():
            param_labels.append(
                parameter['label'] + ' [' + parameter['unit'] + ']'
            )

        self.combobox_parameter.blockSignals(True)
        self.combobox_parameter.clear()
        self.combobox_parameter.addItems(param_labels)
        self.combobox_parameter.blockSignals(False)

    def on_select_parameter(self):
        self.refresh_plot()

    def on_scale_changed(self):
        autoscale = self.autoscale.isChecked()
        self.value_min.setDisabled(autoscale)
        self.value_max.setDisabled(autoscale)
        self.refresh_plot()

    def evaluate(self, blocking=False):
        # If the evaluation is already running, we abort it and reset
        #  the button label
        if self.evaluation_running:
            self.evaluation_abort.value = True
            self.refresh_ui()
            return

        self.evaluation_abort.value = False
        self.evaluation_running = True
        self.button_evaluate.setText('Cancel')
        # While the evaluation is running, we
        # disable switching to multi-peak fit and adjusting bounds
        self.nrBrillouinPeaksGroup.setEnabled(False)
        self.bounds_table.setEnabled(False)
        self.evaluation_timer.start(500)

        self.plot_count = 0
        self.count = mp.Value('I', 0, lock=True)

        # We have to initialize the value correctly here,
        # otherwise the evaluation might abort immediately
        # if max_count is set only after the evaluation_timer
        # triggered for the first time
        image_keys = self.session.get_image_keys()
        self.max_count = mp.Value('i', len(image_keys), lock=True)

        dnkw = {
            "count": self.count,
            "max_count": self.max_count,
            "abort": self.evaluation_abort,
        }

        self.thread = QThread()
        self.worker = Worker(fkw=dnkw)
        self.worker.moveToThread(self.thread)
        self.thread.started.connect(self.worker.run)
        self.worker.finished.connect(self.thread.quit)
        self.worker.finished.connect(self.worker.deleteLater)
        self.worker.finished.connect(self.refresh_ui)
        self.thread.finished.connect(self.thread.deleteLater)
        self.thread.start()

        if blocking:
            while self.evaluation_running:
                QCoreApplication.instance().processEvents()
                time.sleep(0.1)

    def refresh_ui(self):
        # If evaluation is aborted by user,
        # couldn't start or is finished,
        # we stop the timer
        if self.evaluation_abort.value or\
           self.max_count.value < 0 or\
           self.count.value >= self.max_count.value:
            self.evaluation_timer.stop()
            self.evaluation_running = False
            self.button_evaluate.setText('Evaluate')
            self.nrBrillouinPeaksGroup.setEnabled(True)
            session = Session.get_instance()
            if session.evaluation_model().nr_brillouin_peaks > 1:
                self.bounds_table.setEnabled(True)
            self.refresh_plot()

        if self.max_count.value >= 0:
            self.evaluation_progress.setMaximum(self.max_count.value)
        self.evaluation_progress.setValue(self.count.value)

        # We refresh the image every thirty points to not slow down to much
        if (self.count.value - self.plot_count) > 30:
            self.plot_count = self.count.value
            self.refresh_plot()

    def refresh_plot(self):
        session = Session.get_instance()
        evm = session.evaluation_model()
        if evm is None:
            return

        parameters = evm.get_parameter_keys()

        parameter_index = self.combobox_parameter.currentIndex()
        brillouin_peak_index = self.combobox_peak_number.currentIndex()
        parameter_key = list(parameters.keys())[parameter_index]

        data, positions, dimensionality, labels =\
            self.evaluation_controller.\
            get_data(parameter_key, brillouin_peak_index)

        # Subtract the mean value of the positions,
        # so they are centered around zero
        for position in positions:
            position -= np.nanmean(position)

        # Check that we have the correct subplot type
        if dimensionality != 3\
                and isinstance(self.plot, Axes3D):
            self.mplcanvas.get_figure().delaxes(self.plot)
            self.plot = self.mplcanvas.\
                get_figure().add_subplot(111)
        if dimensionality == 3:
            self.mplcanvas.get_figure().clf()
            self.plot = self.mplcanvas.\
                get_figure().add_subplot(111, projection='3d')

        # Create the slices list
        dslice = [slice(None) if dim > 1 else 0 for dim in data.shape]
        idx = [idx for idx, dim in enumerate(data.shape) if dim > 1]

        try:
            if dimensionality == 0:
                # If this is a line plot already, just set new data
                if isinstance(self.image_map, list) and\
                        isinstance(self.image_map[0], matplotlib.lines.Line2D):
                    self.image_map[0].set_data(0, data[tuple(dslice)])
                else:
                    self.clear_plots()
                    self.image_map =\
                        self.plot.plot(0, data[tuple(dslice)], marker='x')
                self.plot.set_xlabel('')
                self.plot.set_title(parameters[parameter_key]['label'])
                ylabel = parameters[parameter_key]['symbol'] +\
                    ' [' + parameters[parameter_key]['unit'] + ']'
                self.plot.set_ylabel(ylabel)
                self.plot.axis('auto')
                self.plot.set_ylim(
                    tuple(
                        data[tuple(dslice)] * np.array([0.99, 1.01])
                    )
                )
            if dimensionality == 1:
                # If this is a line plot already, just set new data
                if isinstance(self.image_map, list) and\
                        isinstance(self.image_map[0], matplotlib.lines.Line2D):
                    self.image_map[0].set_data(
                        positions[idx[0]][tuple(dslice)],
                        data[tuple(dslice)]
                    )
                else:
                    self.clear_plots()
                    self.image_map = self.plot.plot(
                        positions[idx[0]][tuple(dslice)],
                        data[tuple(dslice)]
                    )
                self.plot.set_title(parameters[parameter_key]['label'])
                self.plot.set_xlabel(labels[idx[0]])
                ylabel = parameters[parameter_key]['symbol'] +\
                    ' [' + parameters[parameter_key]['unit'] + ']'
                self.plot.set_ylabel(ylabel)
                self.plot.axis('auto')
                minx = np.nanmin(positions[idx[0]][tuple(dslice)])
                maxx = np.nanmax(positions[idx[0]][tuple(dslice)])
                if minx < maxx:
                    self.plot.set_xlim(minx, maxx)
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        action='ignore',
                        message='All-NaN slice encountered'
                    )
                    (value_min, value_max) = self.get_plot_limits(data)
                    if value_min < value_max:
                        self.plot.set_ylim(
                            value_min,
                            value_max
                        )
            if dimensionality == 2:
                # We rotate the array so the x-axis is shown as the
                # horizontal axis
                image_map = data[tuple(dslice)]
                image_map = np.rot90(image_map)
                extent = np.nanmin(positions[idx[0]][tuple(dslice)]), \
                    np.nanmax(positions[idx[0]][tuple(dslice)]), \
                    np.nanmin(positions[idx[1]][tuple(dslice)]), \
                    np.nanmax(positions[idx[1]][tuple(dslice)])
                if isinstance(self.image_map, matplotlib.image.AxesImage):
                    self.image_map.set_data(image_map)
                    self.image_map.set_extent(extent)
                else:
                    self.clear_plots()
                    self.image_map = self.plot.imshow(
                        image_map, interpolation='nearest',
                        extent=extent
                    )
                    self.colorbar =\
                        self.mplcanvas.get_figure().colorbar(self.image_map)

                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        action='ignore',
                        message='All-NaN slice encountered'
                    )
                    (value_min, value_max) = self.get_plot_limits(data)
                    if value_min < value_max:
                        self.image_map.set_clim(value_min, value_max)
                        # For some reason we have to apply the color limits
                        # twice to make it work properly
                        self.image_map.set_clim(value_min, value_max)
                self.plot.set_title(parameters[parameter_key]['label'])
                self.plot.set_xlabel(labels[idx[0]])
                self.plot.set_ylabel(labels[idx[1]])
                cb_label = parameters[parameter_key]['symbol'] +\
                    ' [' + parameters[parameter_key]['unit'] + ']'
                self.colorbar.ax.set_title(cb_label)
                if self.aspect_ratio.isChecked():
                    self.plot.axis('scaled')
                else:
                    self.plot.axis('auto')
                self.plot.set_xlim(
                    np.nanmin(positions[idx[0]][tuple(dslice)]),
                    np.nanmax(positions[idx[0]][tuple(dslice)])
                )
                self.plot.set_ylim(
                    np.nanmin(positions[idx[1]][tuple(dslice)]),
                    np.nanmax(positions[idx[1]][tuple(dslice)])
                )
            if dimensionality == 3:
                (value_min, value_max) = self.get_plot_limits(data)

                scalar_map = matplotlib.cm.ScalarMappable(
                    norm=Normalize(vmin=value_min, vmax=value_max),
                    cmap=matplotlib.cm.viridis
                )

                plots = []

                # We slice the data along the last occurrence
                # of the shortest dimension
                b = data.shape[::-1]
                axis = len(b) - np.argmin(b) - 1

                for slice_idx in range(data.shape[axis]):
                    dslice[axis] = slice_idx

                    idx_t = tuple(dslice)
                    s = self.plot.plot_surface(
                        positions[0][idx_t],
                        positions[1][idx_t],
                        positions[2][idx_t],
                        facecolors=scalar_map.to_rgba(data[idx_t]),
                        shade=False
                    )
                    plots.append(s)
                self.image_map = plots
                self.plot.set_xlabel(labels[0])
                self.plot.set_ylabel(labels[1])
                self.plot.set_zlabel(labels[2])

                self.colorbar =\
                    self.mplcanvas.get_figure().colorbar(
                        scalar_map,
                        ax=self.plot
                    )
                cb_label = parameters[parameter_key]['symbol'] +\
                    ' [' + parameters[parameter_key]['unit'] + ']'
                self.colorbar.ax.set_title(cb_label)

            self.mplcanvas.draw()
        except Exception as e:
            self.reset_ui()
            raise e

    def get_plot_limits(self, data):
        if self.autoscale.isChecked():
            value_min = np.nanmin(data)
            value_max = np.nanmax(data)
            self.value_min.blockSignals(True)
            self.value_min.setValue(value_min)
            self.value_min.blockSignals(False)
            self.value_max.blockSignals(True)
            self.value_max.setValue(value_max)
            self.value_max.blockSignals(False)
        else:
            value_min = self.value_min.value()
            value_max = self.value_max.value()
            if value_min > value_max:
                value_min = np.nanmin(data)
                value_max = np.nanmax(data)
        if value_min < value_max:
            # Adjust the double spin box step size
            single_step = (value_max - value_min) / 15
            self.value_min.setSingleStep(single_step)
            self.value_max.setSingleStep(single_step)

        return value_min, value_max



import pathlib

from PyQt6 import QtCore
import pytest

from bmlab.session import Session

from bmicro.gui.main import BMicro


def data_file_path(file_name):
    return pathlib.Path(__file__).parent.parent / 'data' / file_name


@pytest.fixture
def window(mocker):
    window = BMicro()
    file_name = data_file_path('Water.h5')

    def mock_getOpenFileName(self, *args, **kwargs):
        return file_name, None

    mocker.patch('PyQt6.QtWidgets.QFileDialog.getOpenFileName',
                 mock_getOpenFileName)
    window.open_file()
    yield window
    window.close()


def test_clicking_rotate_updates_session(qtbot, window):

    session = Session.get_instance()
    assert session.orientation.rotation == 0
    qtbot.mouseClick(
        window.widget_data_view.radio_rotation_90_cw,
        QtCore.Qt.MouseButton.LeftButton)
    assert session.orientation.rotation == 1
    qtbot.mouseClick(
        window.widget_data_view.radio_rotation_90_ccw,
        QtCore.Qt.MouseButton.LeftButton)
    assert session.orientation.rotation == 3

    # For some reason, the following code three lines of code do not trigger
    # the event of the radio_rotation_none radio button. But
    # this happens only for Windows, not for Mac or Linux!
    # This appears to be a bug in the fixture qtbot of pytest-qt,
    # because performing the test "manually" works fine even
    # for Windows.
    #
    # qtbot.mouseClick(
    #     window.widget_data_view.radio_rotation_none, QtCore.Qt.LeftButton)
    # assert window.session.rotation == 0


def test_clicking_reflect_updates_session(qtbot, window):

    session = Session.get_instance()
    assert session.orientation.reflection == {
        'vertically': False, 'horizontally': False}

    qtbot.mouseClick(
        window.widget_data_view.checkbox_reflect_horizontally,
        QtCore.Qt.MouseButton.LeftButton)
    assert session.orientation.reflection == {
        'vertically': False, 'horizontally': True}

    qtbot.mouseClick(
        window.widget_data_view.checkbox_reflect_vertically,
        QtCore.Qt.MouseButton.LeftButton)
    assert session.orientation.reflection == {
        'vertically': True, 'horizontally': True}

    qtbot.mouseClick(
        window.widget_data_view.checkbox_reflect_horizontally,
        QtCore.Qt.MouseButton.LeftButton)
    assert session.orientation.reflection == {
        'vertically': True, 'horizontally': False}


def test_open_file_shows_preview(qtbot, window):
    assert len(window.widget_data_view.mplcanvas.fig.get_axes()) > 0


def test_selecting_setup_updates_session(qtbot, window):
    qtbot.keyClicks(window.widget_data_view.combobox_setup,
                    '532 nm @ Biotec R314')

    assert Session.get_instance().setup.name == '532 nm @ Biotec R314'


def test_close_file_clears_datatab(qtbot, window):
    assert window.widget_data_view.label_selected_file.text() != ''
    assert Session.get_instance().file
    window.close_file()
    assert not Session.get_instance().file
    assert window.widget_data_view.label_selected_file.text() == ''



from PyQt6 import QtWidgets
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT


class MplCanvas(FigureCanvasQTAgg):

    def __init__(self, parent, toolbar=False, width=5, height=5, dpi=100):
        """
        A custom PyQt widget for plotting with matplotlib.

        Parameters
        ----------
        parent: QWidget
            the widget where to embed the current plot widget
        toolbar: tuple of str
            Strings defining what buttons to show on the toolbar
        width: int
            width of the figure
        height: int
            height of the figure
        dpi: int
            DPI settings
        """

        self.fig = Figure(figsize=(width, height), dpi=dpi)
        super(MplCanvas, self).__init__(self.fig)

        layout = QtWidgets.QVBoxLayout()

        if toolbar:

            class CustomToolbar(NavigationToolbar2QT):
                toolitems = [t for t in NavigationToolbar2QT.toolitems if
                             t[0] in toolbar]

            layout.addWidget(CustomToolbar(self, parent))

        layout.addWidget(self)

        parent.setLayout(layout)

    def get_figure(self):
        return self.fig



import pathlib
from importlib import resources
import hashlib
import signal
import sys
import traceback

import numpy as np

from PyQt6 import QtWidgets, uic, QtCore, QtGui
from PyQt6.QtWidgets import QFileDialog, QMessageBox, \
    QVBoxLayout, QWidget, QCheckBox, QHBoxLayout, QLabel, QLineEdit
from PyQt6.QtCore import QSize

from bmlab.session import Session
from bmlab.file import is_source_file
from bmlab.models.setup import AVAILABLE_SETUPS
from bmlab.models import EvaluationModel
from bmlab.controllers import PeakSelectionController, ExportController, \
    EvaluationController

from . import data
from . import extraction
from . import calibration
from . import peak_selection
from . import evaluation

from bmicro import __version__ as bmicroversion
from bmlab import __version__ as bmlabversion


def check_event_mime_data(event):
    """ Returns the path to local file if h5 file """
    if event.mimeData().hasUrls():
        urls = event.mimeData().urls()
        if urls:
            path = urls[0].toLocalFile()
            if path.endswith(".h5"):
                return path
    return False


class BMicro(QtWidgets.QMainWindow):
    """
    Class for the main window of BMicro.
    The application can be started from console by running

        python -m bmicro

    """

    def __init__(self, *args, **kwargs):
        # Settings are stored in the .ini file format. Even though
        # `self.settings` may return integer/bool in the same session,
        # in the next session, it will reliably return strings. Lists
        # of strings (comma-separated) work nicely though.
        QtCore.QCoreApplication.setOrganizationName("BMicro")
        QtCore.QCoreApplication.setOrganizationDomain(
            "bmicro.readthedocs.io")
        QtCore.QCoreApplication.setApplicationName("BMicro")
        QtCore.QSettings.setDefaultFormat(QtCore.QSettings.Format.IniFormat)
        # Some promoted widgets may need the above constants set in order
        # to access the settings upon initialization.

        """ Initializes BMicro."""
        super(BMicro, self).__init__(*args, **kwargs)
        ref = resources.files('bmicro.gui') / 'main.ui'
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self)
        QtCore.QCoreApplication.setApplicationName('BMicro')

        ref = resources.files('bmicro') / 'img'
        with resources.as_file(ref) as imdir:
            self.imdir = imdir

        self.tabWidget.currentChanged.connect(self.update_ui)

        self.export_dialog = None
        # Initialize the export configuration
        self.export_config = ExportController.get_configuration()

        self.batch_dialog = None
        self.batch_files = {}
        self.batch_config = {
            'setup': {
                'set': False,
                'setup': AVAILABLE_SETUPS[0],
            },
            'orientation': {
                'set': False,
                'rotation': 0,
                'reflection': {'vertically': False, 'horizontally': True},
            },
            'extraction': {
                'extract': False,
            },
            'calibration': {
                'find-peaks': False,
                'calibrate': False,
            },
            'peak-selection': {
                'select': False,
                'brillouin_regions': [(4.0e9, 6.0e9), (9.0e9, 11.0e9)],
                'rayleigh_regions': [(-2.0e9, 2.0e9), (13.0e9, 17.0e9)],
            },
            'evaluation': {
                'evaluate': False,
                'nr_brillouin_peaks': 1,
                'bounds_w0': None,
                'bounds_fwhm': None,
            },
            'export': {
                'export': False,
            },
        }
        self.batch_evaluation_running = False

        # Build tabs
        self.widget_data_view = data.DataView(self)
        self.layout_data = QtWidgets.QVBoxLayout()
        self.tab_data.setLayout(self.layout_data)
        self.layout_data.addWidget(self.widget_data_view)
        self.widget_extraction_view = extraction.ExtractionView(self)
        self.layout_extraction = QtWidgets.QVBoxLayout()
        self.tab_extraction.setLayout(self.layout_extraction)
        self.layout_extraction.addWidget(self.widget_extraction_view)
        self.widget_calibration_view = calibration.CalibrationView(self)
        self.layout_calibration = QtWidgets.QVBoxLayout()
        self.tab_calibration.setLayout(self.layout_calibration)
        self.layout_calibration.addWidget(self.widget_calibration_view)
        self.widget_peak_selection_view = peak_selection.PeakSelectionView(
            self)
        self.layout_peak_selection = QtWidgets.QVBoxLayout()
        self.tab_peak_selection.setLayout(self.layout_peak_selection)
        self.layout_peak_selection.addWidget(self.widget_peak_selection_view)
        self.widget_evaluation_view = evaluation.EvaluationView(self)
        self.layout_evaluation = QtWidgets.QVBoxLayout()
        self.tab_evaluation.setLayout(self.layout_evaluation)
        self.layout_evaluation.addWidget(self.widget_evaluation_view)

        self.connect_menu()

        # Connect drag and drop
        self.dragEnterEvent = self.drag_enter_event
        self.dropEvent = self.drop_event

        self.setAcceptDrops(True)

        self.reset_ui()

        self.settings = QtCore.QSettings()

    def connect_menu(self):
        """ Registers the menu actions """
        self.action_open.triggered.connect(self.open_file)
        self.action_close.triggered.connect(self.close_file)
        self.action_save.triggered.connect(self.save_session)
        self.action_export.triggered.connect(self.on_action_export_file)
        self.action_exit.triggered.connect(self.exit_app)

        self.action_about.triggered.connect(self.on_action_about)
        self.action_batch_evaluation.triggered.connect(
            self.on_action_batch_evaluation)

    def open_file(self, file_name=None):
        """ Show open file dialog and load file. """
        if not file_name:
            file_name, _ = QFileDialog.getOpenFileName(
                self, 'Open File...',
                directory=self.settings.value("path/last-used"),
                filter='*.h5')

        """ file_name is empty if user selects 'Cancel' in FileDialog """
        if not file_name:
            return
        else:
            self.settings.setValue("path/last-used",
                                   str(pathlib.Path(file_name).parent))

        session = Session.get_instance()
        try:
            self.close_file()
            session.set_file(file_name)
        except FileNotFoundError as e:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Icon.Warning)
            msg.setText('Unable to load file:')
            msg.setInformativeText(e.strerror)
            msg.setWindowTitle(type(e).__name__)
            msg.exec()
        except Exception as e:
            msg = QMessageBox()
            msg.setIcon(QMessageBox.Icon.Warning)
            msg.setText('An unknown error occured')
            msg.setInformativeText(str(e))
            msg.setWindowTitle('Unknown Error')
            msg.exec()

        self.update_ui()

    def close_file(self):
        Session.get_instance().clear()
        self.reset_ui()

    def on_action_export_file(self):
        self.export_dialog = QtWidgets.QDialog(
            self,
            QtCore.Qt.WindowType.WindowTitleHint |
            QtCore.Qt.WindowType.WindowCloseButtonHint
        )
        ref = resources.files('bmicro.gui') / 'export_configuration.ui'
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self.export_dialog)
        self.export_dialog.setWindowTitle('Export configuration')
        self.export_dialog.setWindowModality(
            QtCore.Qt.WindowModality.ApplicationModal)
        self.export_dialog.button_export.clicked.connect(
            self.export_file
        )
        self.export_dialog.button_cancel.clicked.connect(
            self.close_export_dialog
        )
        self.init_export_dialog(self.export_dialog.widget)
        self.export_dialog.resize(QSize(500, 560))

        self.export_dialog.open()

    def on_export_checkbox(self, parameter, min_box, max_box):
        checked = self.sender().isChecked()
        if checked:
            self.export_config['brillouin']['parameters'].append(parameter)
        else:
            self.export_config['brillouin']['parameters'].remove(parameter)
        min_box.setEnabled(checked)
        max_box.setEnabled(checked)

    def on_export_minbox(self, value, parameter):
        cax = [value, 'max']
        # We use the existing upper limit if available
        if parameter in self.export_config['brillouin']:
            cax[1] = self.export_config['brillouin'][parameter]['cax'][1]
            self.export_config['brillouin'][parameter]['cax'] = tuple(cax)
        else:
            self.export_config['brillouin'][parameter] = {'cax': tuple(cax)}

    def on_export_maxbox(self, value, parameter):
        cax = ['min', value]
        # We use the existing lower limit if available
        if parameter in self.export_config['brillouin']:
            cax[0] = self.export_config['brillouin'][parameter]['cax'][0]
            self.export_config['brillouin'][parameter]['cax'] = tuple(cax)
        else:
            self.export_config['brillouin'][parameter] = {'cax': tuple(cax)}

    def init_export_dialog(self, parent_widget):
        v_layout = QVBoxLayout()
        v_layout.setContentsMargins(0, 6, 0, 6)
        v_layout.setSpacing(0)

        parameters = EvaluationModel.get_default_parameters()
        for key, parameter in parameters.items():
            h_layout = QHBoxLayout()

            # Checkbox
            checkbox = QCheckBox()
            checkbox.setText(
                parameter['label'] + ' [' + parameter['unit'] + ']')
            if key in self.export_config['brillouin']['parameters']:
                checkbox.setChecked(True)
            h_layout.addWidget(checkbox)
            # Spacer
            h_layout.addStretch()

            # Min label and input box
            min_label = QLabel()
            min_label.setText('min')
            h_layout.addWidget(min_label)
            min_box = QLineEdit()
            min_box.setMaximumSize(QSize(50, 20))
            min_box.setMinimumSize(QSize(50, 20))
            h_layout.addWidget(min_box)

            # Max label and input box
            max_label = QLabel()
            max_label.setText('max')
            h_layout.addWidget(max_label)
            max_box = QLineEdit()
            max_box.setMaximumSize(QSize(50, 20))
            max_box.setMinimumSize(QSize(50, 20))
            h_layout.addWidget(max_box)

            # Widget to add the elements to
            parameter_widget = QWidget()
            parameter_widget.setMaximumSize(QSize(5000, 35))
            parameter_widget.setLayout(h_layout)
            v_layout.addWidget(parameter_widget)

            # Set the current config values
            if key in self.export_config['brillouin']:
                min_box.setText(self.export_config['brillouin'][key]['cax'][0])
                max_box.setText(self.export_config['brillouin'][key]['cax'][1])
            # Otherwise we set it to the default min/max
            else:
                min_box.setText('min')
                max_box.setText('max')

            if key in self.export_config['brillouin']['parameters']:
                checkbox.setChecked(True)
                min_box.setEnabled(True)
                max_box.setEnabled(True)
            else:
                min_box.setEnabled(False)
                max_box.setEnabled(False)

            # Connect handlers
            checkbox.clicked.connect(
                lambda checked, param=key, minbox=min_box, maxbox=max_box:
                self.on_export_checkbox(param, minbox, maxbox)
            )
            min_box.textChanged.connect(
                lambda value, param=key: self.on_export_minbox(value, param)
            )
            max_box.textChanged.connect(
                lambda value, param=key: self.on_export_maxbox(value, param)
            )

        # This only works if there is no layout set yet!
        parent_widget.setLayout(v_layout)
        # self.export_dialog.scrollAreaWidgetContents.\
        #     setMinimumSize(QSize(0, 40*len(parameters)))

    def close_export_dialog(self):
        self.export_dialog.close()

    def export_file(self):
        ExportController().export(self.export_config)

    def reset_ui(self):
        """
        Resets the UI if a file is closed.
        """
        self.widget_data_view.reset_ui()
        self.widget_extraction_view.reset_ui()
        self.widget_calibration_view.reset_ui()
        self.widget_peak_selection_view.reset_ui()
        self.widget_evaluation_view.reset_ui()

    def update_ui(self, new_tab_index=-1):
        # If no tab index is specified, we update all tabs
        # (e.g. when a new file is opened).
        # Otherwise, we only update the newly selected tab.
        if new_tab_index == -1:
            self.widget_data_view.update_ui()
            self.widget_extraction_view.update_ui()
            self.widget_calibration_view.update_ui()
            self.widget_peak_selection_view.update_ui()
            self.widget_evaluation_view.update_ui()
        elif new_tab_index == 0:
            self.widget_data_view.update_ui()
        elif new_tab_index == 1:
            self.widget_extraction_view.update_ui()
        elif new_tab_index == 2:
            self.widget_calibration_view.update_ui()
        elif new_tab_index == 3:
            self.widget_peak_selection_view.update_ui()
        elif new_tab_index == 4:
            self.widget_evaluation_view.update_ui()

    @staticmethod
    def drag_enter_event(event):
        """ Handles dragging a file over the GUI """
        if check_event_mime_data(event):
            event.accept()
        else:
            event.ignore()

    def drop_event(self, event):
        """ Handles dropping a file, opens if h5"""
        path = check_event_mime_data(event)
        if path:
            self.open_file(path)

    @staticmethod
    def save_session():
        """ Save the session data to file with ending '.bms' """
        Session.get_instance().save()

    @staticmethod
    def exit_app():
        QtCore.QCoreApplication.quit()

    def on_action_about(self):
        gh = "BrillouinMicroscopy/BMicro"
        rtd = "bmicro.readthedocs.io"
        about_text = f"BMicro {bmicroversion}<br><br>" \
            + "BMicro is a graphical user interface " \
            + "for the analysis of Brillouin microscopy data.<br><br>" \
            + f"Using bmlab {bmlabversion}<br><br>" \
            + "Author: Raimund Schlüßler and others<br>" \
            + "GitHub: " \
            + "<a href='https://github.com/{gh}'>{gh}</a><br>".format(gh=gh) \
            + "Documentation: " \
            + "<a href='https://{rtd}'>{rtd}</a><br>".format(rtd=rtd)
        QtWidgets.QMessageBox.about(self,
                                    f"BMicro {bmicroversion}", about_text)

    def on_action_batch_evaluation(self):
        self.batch_dialog = QtWidgets.QDialog(
            self,
            QtCore.Qt.WindowType.WindowTitleHint |
            QtCore.Qt.WindowType.WindowCloseButtonHint |
            QtCore.Qt.WindowType.WindowMaximizeButtonHint |
            QtCore.Qt.WindowType.WindowMinimizeButtonHint
        )
        ref = resources.files('bmicro.gui') / 'batch_evaluation.ui'
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self.batch_dialog)
        self.batch_dialog.setWindowTitle('Batch evaluation')
        self.batch_dialog.setWindowModality(
            QtCore.Qt.WindowModality.ApplicationModal)
        self.batch_dialog.button_start_cancel.clicked.connect(
            self.start_batch_evaluation
        )
        self.batch_dialog.button_add_folder.clicked.connect(
            self.batch_add_files
        )
        self.batch_dialog.button_remove_folder.clicked.connect(
            self.batch_remove_files
        )
        self.update_batch_file_table()
        self.init_export_dialog(self.batch_dialog.widget_parameters)
        self.batch_dialog.adjustSize()

        self.update_batch_file_settings()

        self.batch_dialog.exec()

    def close_batch_dialog(self):
        self.batch_dialog.close()

    def update_batch_file_settings(self):
        # Setup
        cfg_setup = self.batch_config['setup']
        self.batch_dialog.checkBox_setup.setChecked(cfg_setup['set'])
        self.batch_dialog.checkBox_setup.clicked.connect(self.on_setup_set)
        self.batch_dialog.combobox_setup.addItems(
            [s.name for s in AVAILABLE_SETUPS])
        idx = AVAILABLE_SETUPS.index(cfg_setup['setup'])
        if idx:
            self.batch_dialog.combobox_setup.setCurrentIndex(idx)
        self.batch_dialog.combobox_setup.setEnabled(cfg_setup['set'])
        self.batch_dialog.combobox_setup.currentIndexChanged.connect(
            self.on_setup_select)

        # Orientation
        cfg_orientation = self.batch_config['orientation']
        self.batch_dialog.checkBox_orientation\
            .setChecked(cfg_orientation['set'])
        self.batch_dialog.checkBox_orientation\
            .clicked.connect(self.on_orientation_set)
        self.batch_dialog.groupBox_rotation\
            .setEnabled(cfg_orientation['set'])
        self.batch_dialog.groupBox_reflection\
            .setEnabled(cfg_orientation['set'])

        self.batch_dialog.checkbox_reflect_vertically.setChecked(
            cfg_orientation['reflection']['vertically'])
        self.batch_dialog.checkbox_reflect_horizontally.setChecked(
            cfg_orientation['reflection']['horizontally'])

        self.batch_dialog.radio_rotation_none.setChecked(
            cfg_orientation['rotation'] % 4 == 0)
        self.batch_dialog.radio_rotation_90_cw.setChecked(
            cfg_orientation['rotation'] % 4 == 1)
        self.batch_dialog.radio_rotation_90_ccw.setChecked(
            cfg_orientation['rotation'] % 4 == 3)

        self.batch_dialog.radio_rotation_none\
            .clicked.connect(self.on_rotation_clicked)
        self.batch_dialog.radio_rotation_90_cw\
            .clicked.connect(self.on_rotation_clicked)
        self.batch_dialog.radio_rotation_90_ccw\
            .clicked.connect(self.on_rotation_clicked)

        self.batch_dialog.checkbox_reflect_vertically.toggled.connect(
            self.on_reflection_clicked)
        self.batch_dialog.checkbox_reflect_horizontally.toggled.connect(
            self.on_reflection_clicked)

        # Extraction
        cfg_extraction = self.batch_config['extraction']
        self.batch_dialog.checkBox_extraction\
            .setChecked(cfg_extraction['extract'])
        self.batch_dialog.checkBox_extraction\
            .clicked.connect(self.on_extraction_set)

        # Calibration
        cfg_calibration = self.batch_config['calibration']
        self.batch_dialog.checkBox_find_peaks\
            .setChecked(cfg_calibration['find-peaks'])
        self.batch_dialog.checkBox_find_peaks\
            .clicked.connect(self.on_calibration_find_peaks)
        self.batch_dialog.checkBox_calibrate\
            .setChecked(cfg_calibration['calibrate'])
        self.batch_dialog.checkBox_calibrate\
            .clicked.connect(self.on_calibration_calibrate)

        self.batch_dialog.temperature.valueChanged.connect(
            self.temperature_changed
        )
        self.batch_dialog.shift_0.valueChanged.connect(
            self.shift_methanol_changed
        )
        self.batch_dialog.shift_1.valueChanged.connect(
            self.shift_water_changed
        )

        self.update_batch_calibration_frequencies()

        # Peak selection
        cfg_peak_selection = self.batch_config['peak-selection']
        self.batch_dialog.checkBox_peak_selection\
            .setChecked(cfg_peak_selection['select'])
        self.batch_dialog.checkBox_peak_selection\
            .clicked.connect(self.on_peak_selection_select)
        self.batch_dialog.tableWidget_Brillouin\
            .setEnabled(cfg_peak_selection['select'])
        self.batch_dialog.tableWidget_Rayleigh\
            .setEnabled(cfg_peak_selection['select'])
        self.batch_dialog.button_region_brillouin_remove.clicked.connect(
            lambda: self.remove_evaluation_regions(
                self.batch_dialog.tableWidget_Brillouin,
                self.batch_config[
                    'peak-selection']['brillouin_regions']
            )
        )
        self.batch_dialog.button_region_brillouin_add.clicked.connect(
            lambda: self.add_evaluation_regions(
                self.batch_dialog.tableWidget_Brillouin,
                self.batch_config[
                    'peak-selection']['brillouin_regions']
            )
        )
        self.batch_dialog.button_region_rayleigh_remove.clicked.connect(
            lambda: self.remove_evaluation_regions(
                self.batch_dialog.tableWidget_Rayleigh,
                self.batch_config[
                    'peak-selection']['rayleigh_regions']
            )
        )
        self.batch_dialog.button_region_rayleigh_add.clicked.connect(
            lambda: self.add_evaluation_regions(
                self.batch_dialog.tableWidget_Rayleigh,
                self.batch_config[
                    'peak-selection']['rayleigh_regions']
            )
        )

        self.update_evaluation_regions_tables()

        self.batch_dialog.tableWidget_Brillouin.itemChanged.connect(
            lambda item: self.on_region_changed(
                self.batch_config['peak-selection']['brillouin_regions'],
                item)
        )

        self.batch_dialog.tableWidget_Rayleigh.itemChanged.connect(
            lambda item: self.on_region_changed(
                self.batch_config['peak-selection']['rayleigh_regions'],
                item)
        )

        header = self.batch_dialog.tableWidget_Brillouin.horizontalHeader()
        header.setSectionResizeMode(0,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.batch_dialog.tableWidget_Brillouin\
            .verticalHeader().setVisible(False)
        self.batch_dialog.tableWidget_Brillouin \
            .setHorizontalHeaderLabels(["start", "end"])

        header = self.batch_dialog.tableWidget_Rayleigh.horizontalHeader()
        header.setSectionResizeMode(0,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.batch_dialog.tableWidget_Rayleigh\
            .verticalHeader().setVisible(False)
        self.batch_dialog.tableWidget_Rayleigh\
            .setHorizontalHeaderLabels(["start", "end"])

        # Evaluation
        cfg_evaluation = self.batch_config['evaluation']
        self.batch_dialog.checkBox_evaluation\
            .setChecked(cfg_evaluation['evaluate'])
        self.batch_dialog.checkBox_evaluation\
            .clicked.connect(self.on_evaluation_evaluate)
        self.batch_dialog.nrBrillouinPeaksGroup\
            .setEnabled(cfg_evaluation['evaluate'])
        if cfg_evaluation['nr_brillouin_peaks'] == 1:
            self.batch_dialog.nrBrillouinPeaks_1.setChecked(True)
        elif cfg_evaluation['nr_brillouin_peaks'] == 2:
            self.batch_dialog.nrBrillouinPeaks_2.setChecked(True)
        elif cfg_evaluation['nr_brillouin_peaks'] == 4:
            self.batch_dialog.nrBrillouinPeaks_4.setChecked(True)
        self.batch_dialog.bounds_table\
            .setEnabled(cfg_evaluation['evaluate'] and
                        cfg_evaluation['nr_brillouin_peaks'] > 1)

        self.batch_dialog.bounds_table.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.Stretch)
        self.batch_dialog.bounds_table.verticalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.ResizeMode.Stretch)

        self.update_batch_bounds_table()
        self.batch_dialog.bounds_table.cellChanged.connect(
            self.evaluation_bounds_changed)

        self.batch_dialog.nrBrillouinPeaks_1.toggled.connect(
            lambda: self.set_nr_brillouin_peaks(1))
        self.batch_dialog.nrBrillouinPeaks_2.toggled.connect(
            lambda: self.set_nr_brillouin_peaks(2))
        self.batch_dialog.nrBrillouinPeaks_4.toggled.connect(
            lambda: self.set_nr_brillouin_peaks(4))

        # Export
        cfg_export = self.batch_config['export']
        self.batch_dialog.checkBox_export\
            .setChecked(cfg_export['export'])
        self.batch_dialog.checkBox_export\
            .clicked.connect(self.on_export_export)

    def temperature_changed(self):
        temperature = self.sender().value()

        self.batch_config['setup']['setup'].set_temperature(temperature)
        self.update_batch_calibration_frequencies()

    def shift_methanol_changed(self):
        shift = self.sender().value()

        self.batch_config['setup']['setup']\
            .calibration.set_shift_methanol(1e9 * shift)

    def shift_water_changed(self):
        shift = self.sender().value()

        self.batch_config['setup']['setup']\
            .calibration.set_shift_water(1e9 * shift)

    def update_batch_calibration_frequencies(self):
        cfg_setup = self.batch_config['setup']
        if not cfg_setup['set']:
            self.batch_dialog.shift_0.setEnabled(False)
            self.batch_dialog.label_shift_0.setEnabled(False)
            self.batch_dialog.shift_1.setEnabled(False)
            self.batch_dialog.label_shift_1.setEnabled(False)
            self.batch_dialog.temperature.setEnabled(False)
            self.batch_dialog.label_temperature.setEnabled(False)
        else:
            # Set current calibration values
            if cfg_setup['setup'].calibration.shift_methanol is not None:
                self.batch_dialog.label_shift_0.setEnabled(True)
                self.batch_dialog.shift_0.setEnabled(True)
                self.batch_dialog.shift_0.setValue(
                    1e-9 * cfg_setup['setup'].calibration.shift_methanol)
            else:
                self.batch_dialog.shift_0.setEnabled(False)
                self.batch_dialog.label_shift_0.setEnabled(False)

            if cfg_setup['setup'].calibration.shift_water is not None:
                self.batch_dialog.label_shift_1.setEnabled(True)
                self.batch_dialog.shift_1.setEnabled(True)
                self.batch_dialog.shift_1.setValue(
                    1e-9 * cfg_setup['setup'].calibration.shift_water)
            else:
                self.batch_dialog.shift_1.setEnabled(False)
                self.batch_dialog.label_shift_1.setEnabled(False)

            self.batch_dialog.temperature.setEnabled(True)
            self.batch_dialog.label_temperature.setEnabled(True)
            self.batch_dialog.temperature.setValue(
                cfg_setup['setup'].temperature - 273.15
            )

    def update_batch_bounds_table(self):
        bounds_w0 = self.batch_config['evaluation']['bounds_w0']
        bounds_fwhm = self.batch_config['evaluation']['bounds_fwhm']
        if bounds_w0 is None or bounds_fwhm is None:
            self.batch_dialog.bounds_table.setRowCount(0)
        else:
            self.batch_dialog.bounds_table.setColumnCount(4)
            self.batch_dialog.bounds_table.setRowCount(len(bounds_w0))
            for i, bound in enumerate(bounds_w0):
                item = QtWidgets.QTableWidgetItem(str(bound[0]))
                self.batch_dialog.bounds_table.setItem(i, 0, item)
                item = QtWidgets.QTableWidgetItem(str(bound[1]))
                self.batch_dialog.bounds_table.setItem(i, 1, item)
            for i, bound in enumerate(bounds_fwhm):
                item = QtWidgets.QTableWidgetItem(str(bound[0]))
                self.batch_dialog.bounds_table.setItem(i, 2, item)
                item = QtWidgets.QTableWidgetItem(str(bound[1]))
                self.batch_dialog.bounds_table.setItem(i, 3, item)

        self.batch_dialog.bounds_table.setEnabled(
            self.batch_config['evaluation']['evaluate'] and
            self.batch_config['evaluation']['nr_brillouin_peaks'] > 1)

    def set_nr_brillouin_peaks(self, nr_peaks):
        if not self.sender().isChecked():
            return
        self.batch_config['evaluation']['nr_brillouin_peaks'] = nr_peaks

        if nr_peaks == 1:
            bounds_w0 = None
            bounds_fwhm = None

        # Initialize the bounds if necessary
        if nr_peaks > 1 and\
                (self.batch_config['evaluation']['bounds_w0'] is None or
                 len(self.batch_config['evaluation']['bounds_w0'])
                 is not nr_peaks):
            bounds_w0 = [['min', 'max'] for _ in range(nr_peaks)]
        if nr_peaks > 1 and\
                (self.batch_config['evaluation']['bounds_fwhm'] is None or
                 len(self.batch_config['evaluation']['bounds_fwhm'])
                 is not nr_peaks):
            bounds_fwhm = [['0', 'inf'] for _ in range(nr_peaks)]

        self.batch_config['evaluation']['bounds_w0'] = bounds_w0
        self.batch_config['evaluation']['bounds_fwhm'] = bounds_fwhm

        self.update_batch_bounds_table()

    def evaluation_bounds_changed(self, row, column):
        if column < 2:
            self.batch_config['evaluation']['bounds_w0'][row][column] =\
                    self.batch_dialog.bounds_table.item(row, column).text()
        elif column < 4:
            self.batch_config['evaluation']['bounds_fwhm'][row][column - 2] =\
                    self.batch_dialog.bounds_table.item(row, column).text()

    def add_evaluation_regions(self, region_table, region_list):
        region_list.append((0., 0.))

        self.update_evaluation_regions_tables()

        region_table.setFocus()
        region_table.setCurrentCell(len(region_list) - 1, 0)

    def remove_evaluation_regions(self, region_table, region_list):
        selected_ranges = region_table.selectedRanges()
        sorted_ranges = sorted(
            selected_ranges,
            key=lambda item: item.topRow(),
            reverse=True
        )
        for sorted_range in sorted_ranges:
            start = sorted_range.topRow()
            end = sorted_range.bottomRow()
            for idx in reversed(range(start, end + 1)):
                del region_list[idx]
        region_table.clearSelection()

        self.update_evaluation_regions_tables()

    def update_evaluation_regions_tables(self):
        cfg_peak_selection = self.batch_config['peak-selection']

        brillouin_regions = cfg_peak_selection['brillouin_regions']
        brillouin_table = self.batch_dialog.tableWidget_Brillouin
        brillouin_table.setRowCount(len(brillouin_regions))
        for rowIdx, region in enumerate(brillouin_regions):
            # Add regions to table
            # Block signals, so the itemChanged signal is not
            # emitted during table creation
            brillouin_table.blockSignals(True)
            for columnIdx, value in enumerate(region):
                region_item = QtWidgets.QTableWidgetItem(str(1e-9 * value))
                brillouin_table.setItem(rowIdx, columnIdx, region_item)
            brillouin_table.blockSignals(False)

        rayleigh_regions = cfg_peak_selection['rayleigh_regions']
        rayleigh_table = self.batch_dialog.tableWidget_Rayleigh
        rayleigh_table.setRowCount(len(rayleigh_regions))
        for rowIdx, region in enumerate(rayleigh_regions):
            # Add regions to table
            # Block signals, so the itemChanged signal is not
            # emitted during table creation
            rayleigh_table.blockSignals(True)
            for columnIdx, value in enumerate(region):
                region_item = QtWidgets.QTableWidgetItem(str(1e-9 * value))
                rayleigh_table.setItem(rowIdx, columnIdx, region_item)
            rayleigh_table.blockSignals(False)

    def on_setup_set(self):
        self.batch_config['setup']['set'] = self.sender().isChecked()
        self.batch_dialog.combobox_setup\
            .setEnabled(self.batch_config['setup']['set'])

        self.update_batch_calibration_frequencies()

    def on_orientation_set(self):
        self.batch_config['orientation']['set'] = self.sender().isChecked()
        self.batch_dialog.groupBox_rotation\
            .setEnabled(self.batch_config['orientation']['set'])
        self.batch_dialog.groupBox_reflection\
            .setEnabled(self.batch_config['orientation']['set'])

    def on_rotation_clicked(self):
        """
        Action triggered when user clicks one of the rotation radio buttons.
        """

        radio_button = self.sender()
        if not radio_button.isChecked():
            return

        if radio_button == self.batch_dialog.radio_rotation_none:
            self.batch_config['orientation']['rotation'] = 0
        elif radio_button == self.batch_dialog.radio_rotation_90_cw:
            self.batch_config['orientation']['rotation'] = 1
        elif radio_button == self.batch_dialog.radio_rotation_90_ccw:
            self.batch_config['orientation']['rotation'] = 3

    def on_reflection_clicked(self):
        """ Triggered when a reflection checkbox is clicked """

        checkbox = self.sender()

        if checkbox == self.batch_dialog.checkbox_reflect_vertically:
            self.batch_config['orientation'][
                'reflection']['vertically'] = checkbox.isChecked()
        elif checkbox == self.batch_dialog.checkbox_reflect_horizontally:
            self.batch_config['orientation'][
                'reflection']['horizontally'] = checkbox.isChecked()

    def on_extraction_set(self):
        self.batch_config['extraction']['extract'] = self.sender().isChecked()

    def on_calibration_find_peaks(self):
        self.batch_config['calibration']['find-peaks']\
            = self.sender().isChecked()

    def on_calibration_calibrate(self):
        self.batch_config['calibration']['calibrate']\
            = self.sender().isChecked()

    def on_peak_selection_select(self):
        self.batch_config['peak-selection']['select']\
            = self.sender().isChecked()
        self.batch_dialog.tableWidget_Brillouin\
            .setEnabled(self.batch_config['peak-selection']['select'])
        self.batch_dialog.tableWidget_Rayleigh\
            .setEnabled(self.batch_config['peak-selection']['select'])

    @staticmethod
    def on_region_changed(region_table, item):
        row = item.row()
        column = item.column()
        value = float(item.text())

        current_region = np.asarray(region_table[row])
        current_region[column] = 1e9 * value
        current_region = tuple(current_region)
        region_table[row] = current_region

    def on_evaluation_evaluate(self):
        self.batch_config['evaluation']['evaluate']\
            = self.sender().isChecked()
        self.batch_dialog.nrBrillouinPeaksGroup\
            .setEnabled(self.batch_config['evaluation']['evaluate'])
        self.batch_dialog.bounds_table\
            .setEnabled(
                self.batch_config['evaluation']['evaluate'] and
                self.batch_config['evaluation']['nr_brillouin_peaks'] > 1)

    def on_export_export(self):
        self.batch_config['export']['export'] = self.sender().isChecked()

    def on_setup_select(self):
        """
        Action triggered when the user selects a different setup.
        """
        name = self.batch_dialog.combobox_setup.currentText()
        setup = None
        for s in AVAILABLE_SETUPS:
            if s.name == name:
                setup = s
                break
        self.batch_config['setup']['setup'] = setup

        self.update_batch_calibration_frequencies()

    def start_batch_evaluation(self):
        self.batch_evaluation_running = not self.batch_evaluation_running
        if self.batch_evaluation_running:
            self.run_batch_evaluation()
        else:
            self.widget_evaluation_view.evaluation_abort.value = True

    def run_batch_evaluation(self):
        self.batch_dialog.button_start_cancel.setText('Cancel')
        self.batch_dialog.progressBar.setMaximum(len(self.batch_files))
        self.batch_dialog.progressBar.setValue(0)
        for i, (file_hash, file) in enumerate(self.batch_files.items()):
            try:
                self.evaluate_batch_file(file)
            except BaseException:
                # Set the status as failed
                file['status'] = 'failed'
                self.update_batch_file_table()
            if not self.batch_evaluation_running:
                break
            self.batch_dialog.progressBar.setValue(i + 1)
        self.batch_dialog.button_start_cancel.setText('Start')
        self.batch_evaluation_running = False
        self.update_batch_file_table()

    def evaluate_batch_file(self, file):
        # Set the status as in process
        file['status'] = 'in-process'
        self.update_batch_file_table()

        # Show the data tab and open the file
        self.tabWidget.setCurrentIndex(0)
        QtCore.QCoreApplication.instance().processEvents()
        self.open_file(file['path'])
        QtCore.QCoreApplication.instance().processEvents()

        """
        Evaluate the file
        """
        session = Session.get_instance()

        rep_keys = session.file.repetition_keys()

        for rep_key in rep_keys:
            # Load repetition
            session.set_current_repetition(rep_key)
            QtCore.QCoreApplication.instance().processEvents()

            # Setup
            cfg_setup = self.batch_config['setup']
            if cfg_setup['set']:
                session.set_setup(cfg_setup['setup'])

            # Orientation
            cfg_orientation = self.batch_config['orientation']
            if cfg_orientation['set']:
                session.set_rotation(cfg_orientation['rotation'])
                session.set_reflection(
                    vertically=cfg_orientation['reflection']['vertically'],
                    horizontally=cfg_orientation['reflection']['horizontally']
                )

            # Extraction
            cfg_extraction = self.batch_config['extraction']
            if cfg_extraction['extract']:
                self.tabWidget.setCurrentIndex(1)
                if self.aborted(file):
                    return
                QtCore.QCoreApplication.instance().processEvents()
                self.widget_extraction_view.find_points_all()

            # Calibration
            cfg_calibration = self.batch_config['calibration']
            if cfg_calibration['find-peaks'] or\
                    cfg_calibration['calibrate']:
                self.tabWidget.setCurrentIndex(2)
                if self.aborted(file):
                    return
                QtCore.QCoreApplication.instance().processEvents()
                if not cfg_calibration['find-peaks']:
                    self.widget_calibration_view.\
                        calibrate_all(do_not='find_peaks')
                elif not cfg_calibration['calibrate']:
                    self.widget_calibration_view.\
                        calibrate_all(do_not='calibrate')
                else:
                    self.widget_calibration_view.calibrate_all()

            # PeakSelection
            cfg_peak_selection = self.batch_config['peak-selection']
            if cfg_peak_selection['select']:
                self.tabWidget.setCurrentIndex(3)
                if self.aborted(file):
                    return
                QtCore.QCoreApplication.instance().processEvents()
                psc = PeakSelectionController()
                for brillouin_region \
                        in cfg_peak_selection['brillouin_regions']:
                    psc.add_brillouin_region_frequency(brillouin_region)
                for rayleigh_region \
                        in cfg_peak_selection['rayleigh_regions']:
                    psc.add_rayleigh_region_frequency(rayleigh_region)
                self.widget_peak_selection_view.update_ui()
                QtCore.QCoreApplication.instance().processEvents()

            # Evaluation
            cfg_evaluation = self.batch_config['evaluation']
            if cfg_evaluation['evaluate']:
                self.tabWidget.setCurrentIndex(4)
                if self.aborted(file):
                    return
                QtCore.QCoreApplication.instance().processEvents()
                evc = EvaluationController()
                evc.set_nr_brillouin_peaks(
                    cfg_evaluation['nr_brillouin_peaks'])
                evc.set_bounds(cfg_evaluation['bounds_w0'])
                evc.set_bounds_fwhm(cfg_evaluation['bounds_fwhm'])
                self.widget_evaluation_view.evaluate(blocking=True)

            cfg_export = self.batch_config['export']
            if cfg_export['export']:
                if self.aborted(file):
                    return
                QtCore.QCoreApplication.instance().processEvents()
                self.export_file()

        # Save the evaluated data
        self.save_session()
        if self.aborted(file):
            return
        QtCore.QCoreApplication.instance().processEvents()

        # Close the file
        self.close_file()
        QtCore.QCoreApplication.instance().processEvents()

        # Set the status as done
        file['status'] = 'success'
        self.update_batch_file_table()

    def aborted(self, file):
        if not self.batch_evaluation_running:
            file['status'] = 'aborted'
            self.close_file()
            self.batch_evaluation_running = False
            self.update_batch_file_table()
            self.batch_dialog.progressBar.setValue(0)
            QtCore.QCoreApplication.instance().processEvents()
            return True

    def batch_add_files(self):
        folder_name = QFileDialog.getExistingDirectory(
            self, 'Select folder...',
            directory=self.settings.value("path/last-used")
        )

        if not folder_name:
            return

        # Find all h5 files in the selected folder
        h5_files = pathlib.Path(folder_name).glob('**/*.h5')

        # Add source files to batch if not present yet
        for h5_file in h5_files:
            h5_file_md5 = hashlib.md5(str(h5_file).encode('utf-8')).hexdigest()
            if is_source_file(h5_file)\
                    and h5_file_md5 not in self.batch_files:
                self.batch_files[h5_file_md5] = {
                    'path': h5_file,
                    'status': 'pending',
                }

        self.update_batch_file_table()

    def batch_remove_files(self):
        table = self.batch_dialog.table_files
        selected_ranges = table.selectedRanges()
        hashes = list(self.batch_files.keys())
        hashes_remove = []
        for selected_range in selected_ranges:
            start = selected_range.topRow()
            end = selected_range.bottomRow()
            for i in range(start, end + 1):
                hashes_remove.append(hashes[i])
        for hash_remove in hashes_remove:
            self.batch_files.pop(hash_remove)
        table.clearSelection()
        self.update_batch_file_table()

    def update_batch_file_table(self):
        table = self.batch_dialog.table_files
        table.setTextElideMode(QtCore.Qt.TextElideMode.ElideRight)
        table.setWordWrap(False)
        table.setColumnCount(2)
        table.setRowCount(len(self.batch_files))
        table.blockSignals(True)

        pending_icon_path = str(pathlib.Path(self.imdir) / "pending.svg")
        inprocess_icon_path = str(pathlib.Path(self.imdir) / "in-process.svg")
        success_icon_path =\
            str(pathlib.Path(self.imdir) / "check_circle_outline.svg")
        failed_icon_path = str(pathlib.Path(self.imdir) / "error.svg")
        aborted_icon_path = str(pathlib.Path(self.imdir) / "cancel.svg")

        for rowIdx, (file_hash, file) in enumerate(self.batch_files.items()):
            table.setIconSize(QtCore.QSize(20, 20))
            status = file['status']
            if status == 'in-process':
                icon_path = inprocess_icon_path
            elif status == 'success':
                icon_path = success_icon_path
            elif status == 'aborted':
                icon_path = aborted_icon_path
            elif status == 'failed':
                icon_path = failed_icon_path
            else:
                icon_path = pending_icon_path
            icon = QtGui.QIcon(icon_path)
            entry = QtWidgets.QTableWidgetItem()
            entry.setSizeHint(QtCore.QSize(20, 20))
            entry.setIcon(icon)
            table.setItem(rowIdx, 0, entry)

            path = QtWidgets.QTableWidgetItem(str(file['path']))
            table.setItem(rowIdx, 1, path)
        table.setColumnWidth(0, 18)
        table.horizontalHeader().setStretchLastSection(True)
        table.verticalHeader().setVisible(False)
        table.horizontalHeader().setVisible(False)
        table.blockSignals(False)


def excepthook(etype, value, trace):
    """
    Handler for all unhandled exceptions.
    Parameters
    ----------
    etype: Exception
        the exception type (`SyntaxError`,
        `ZeroDivisionError`, etc...)
    value: str
        the exception error message
    trace: str
        the traceback header, if any (otherwise, it
        prints the standard Python header: ``Traceback (most recent
        call last)``.
    """
    vinfo = "Unhandled exception in BMicro version {}:\n".format(
        bmicroversion)
    tmp = traceback.format_exception(etype, value, trace)
    exception = "".join([vinfo]+tmp)

    errorbox = QtWidgets.QMessageBox()
    errorbox.addButton(QtWidgets.QPushButton('Close'),
                       QtWidgets.QMessageBox.ButtonRole.YesRole)
    errorbox.addButton(QtWidgets.QPushButton(
        'Copy text && Close'), QtWidgets.QMessageBox.ButtonRole.NoRole)
    errorbox.setText(exception)
    ret = errorbox.exec()
    if ret == 1:
        cb = QtWidgets.QApplication.clipboard()
        cb.clear(mode=cb.Mode.Clipboard)
        cb.setText(exception)


# Make Ctr+C close the app
signal.signal(signal.SIGINT, signal.SIG_DFL)
# Display exception hook in separate dialog instead of crashing
sys.excepthook = excepthook



from importlib import resources
import logging

from PyQt6 import uic, QtWidgets
import matplotlib

from bmlab.models.setup import AVAILABLE_SETUPS
from bmlab.session import Session

from bmicro.gui.mpl import MplCanvas


import os

matplotlib.use('Qt5Agg')
logger = logging.getLogger(__name__)


class DataView(QtWidgets.QWidget):
    """
    Class for the data widget
    """

    def __init__(self, *args, **kwargs):
        super(DataView, self).__init__(*args, **kwargs)

        ref = resources.files('bmicro.gui.data') / 'data_view.ui'
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self)

        self.mplcanvas = MplCanvas(self.image_preview_widget)
        self.preview = self.mplcanvas.get_figure().add_subplot(111)
        self.preview.axis('off')

        self.radio_rotation_none.clicked.connect(self.on_rotation_clicked)
        self.radio_rotation_90_cw.clicked.connect(self.on_rotation_clicked)
        self.radio_rotation_90_ccw.clicked.connect(self.on_rotation_clicked)

        self.checkbox_reflect_vertically.toggled.connect(
            self.on_reflection_clicked)
        self.checkbox_reflect_horizontally.toggled.connect(
            self.on_reflection_clicked)

        self.comboBox_repetition.currentIndexChanged.connect(
            self.on_select_repetition)

        self.combobox_setup.addItems([s.name for s in AVAILABLE_SETUPS])
        self.combobox_setup.currentIndexChanged.connect(
            self.on_select_setup)

        # Initialize current setup to first entry
        Session.get_instance().set_setup(AVAILABLE_SETUPS[0])

        self.update_ui()

    def update_ui(self):
        session = Session.get_instance()
        if not session.file:
            return

        filename = str(os.path.basename(session.file.path))
        self.label_selected_file.setText(filename)
        self.label_selected_file.setToolTip(str(session.file.path))
        self.label_selected_file.adjustSize()
        rep_keys = session.file.repetition_keys()
        # Update repetition keys if they have changed
        current_keys = [self.comboBox_repetition.itemText(i)
                        for i in range(self.comboBox_repetition.count())]
        if current_keys != rep_keys:
            self.comboBox_repetition.clear()
            self.comboBox_repetition.addItems(rep_keys)

        if rep_keys and self.comboBox_repetition.currentText():
            repetition = session.current_repetition()
            res = repetition.payload.resolution
            if res:
                self.label_resolution_x.setText(str(res[0]))
                self.label_resolution_y.setText(str(res[1]))
                self.label_resolution_z.setText(str(res[2]))
            date = repetition.date.strftime('%Y-%m-%d %H:%M')
            self.label_date.setText(date)
            self.textedit_comment.setText(session.file.comment)
            self.label_calibration.setText(
                str(not repetition.calibration.is_empty()))

        if session.setup is not None:
            self.combobox_setup.setCurrentText(session.setup.name)

        self.checkbox_reflect_vertically.setChecked(
            session.orientation.reflection['vertically'])
        self.checkbox_reflect_horizontally.setChecked(
            session.orientation.reflection['horizontally'])

        self.radio_rotation_none.setChecked(
            session.orientation.rotation % 4 == 0)
        self.radio_rotation_90_cw.setChecked(
            session.orientation.rotation % 4 == 1)
        self.radio_rotation_90_ccw.setChecked(
            session.orientation.rotation % 4 == 3)

        self.update_preview()

    def reset_ui(self):
        """
        Reset the tab to the state when no file is loaded.
        """
        self.label_selected_file.setText('')
        self.label_selected_file.setToolTip('')
        self.comboBox_repetition.clear()
        self.label_date.setText('')
        self.label_resolution_x.setText('')
        self.label_resolution_y.setText('')
        self.label_resolution_z.setText('')
        self.label_calibration.setText('')
        self.textedit_comment.setText('')
        self.combobox_setup.setCurrentText(AVAILABLE_SETUPS[0].name)
        self.radio_rotation_none.setChecked(True)
        self.checkbox_reflect_vertically.setChecked(False)
        self.checkbox_reflect_horizontally.setChecked(False)
        self.update_preview()

    def on_rotation_clicked(self):
        """
        Action triggered when user clicks one of the rotation radio buttons.
        """

        radio_button = self.sender()
        if not radio_button.isChecked():
            return

        session = Session.get_instance()

        if radio_button == self.radio_rotation_none:
            session.set_rotation(0)
        elif radio_button == self.radio_rotation_90_cw:
            session.set_rotation(1)
        elif radio_button == self.radio_rotation_90_ccw:
            session.set_rotation(3)

        self.update_preview()

    def on_reflection_clicked(self):
        """ Triggered when a reflection checkbox is clicked """

        checkbox = self.sender()

        session = Session.get_instance()
        if checkbox == self.checkbox_reflect_vertically:
            session.set_reflection(vertically=checkbox.isChecked())
        elif checkbox == self.checkbox_reflect_horizontally:
            session.set_reflection(
                horizontally=checkbox.isChecked())

        self.update_preview()

    def update_preview(self):
        """
        Updates the preview plot based on current session settings.
        """
        session = Session.get_instance()
        rep = session.current_repetition()

        if rep and rep.payload.image_keys():
            first_key = rep.payload.image_keys()[0]
            img = session.get_payload_image(first_key, 0)

            self.preview.clear()

            # imshow should always get the transposed image such that
            # the horizontal axis of the plot coincides with the
            # 0-axis of the plotted array:
            self.preview.imshow(img.T, origin='lower', vmin=100, vmax=300)
            self.preview.axis('off')
            self.mplcanvas.draw()
        else:
            self.mplcanvas.get_figure().clf()
            self.preview = self.mplcanvas.get_figure().add_subplot(111)
            self.mplcanvas.draw()

    def on_select_repetition(self):
        """
        Action triggered when the user selects a different repetition.
        """
        session = Session.get_instance()
        rep_key = self.comboBox_repetition.currentText()
        session.set_current_repetition(rep_key)
        self.update_preview()

    def on_select_setup(self):
        """
        Action triggered when the user selects a different setup.
        """
        name = self.combobox_setup.currentText()
        setup = None
        session = Session.get_instance()
        for s in AVAILABLE_SETUPS:
            if s.name == name:
                setup = s
                break
        session.set_setup(setup)



from importlib import resources
import logging

from PyQt6 import QtWidgets, uic

from matplotlib.widgets import SpanSelector
import numpy as np

from bmlab.session import Session
from bmlab.controllers import EvaluationController

from bmicro.gui.mpl import MplCanvas
import warnings

logger = logging.getLogger(__name__)

MODE_DEFAULT = 'default'
MODE_SELECT_BRILLOUIN = 'select_brillouin_peaks'
MODE_SELECT_RAYLEIGH = 'select_rayleigh_peaks'


class PeakSelectionView(QtWidgets.QWidget):
    """
    Class for the peak selection widget
    """

    def __init__(self, *args, **kwargs):
        super(PeakSelectionView, self).__init__(*args, **kwargs)

        ref = (resources.files('bmicro.gui.peak_selection') /
               'peak_selection_view.ui')
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self)

        self.mplcanvas = MplCanvas(self.image_widget,
                                   toolbar=('Home', 'Pan', 'Zoom'))
        self.plot = self.mplcanvas.get_figure().add_subplot(111)

        props = dict(facecolor='green', alpha=0.5)
        self.span_selector = SpanSelector(
            self.plot, onselect=self.on_select_data_region,
            useblit=True,
            direction='horizontal', props=props)

        self.button_brillouin_select_done.clicked.connect(
            self.on_select_brillouin_clicked)
        self.button_rayleigh_select_done.clicked.connect(
            self.on_select_rayleigh_clicked)
        self.button_brillouin_clear.released.connect(
            self.clear_regions)
        self.button_rayleigh_clear.released.connect(
            self.clear_regions)

        self.mode = MODE_DEFAULT

        self.table_Brillouin_regions.itemChanged.connect(
            lambda item: self.on_region_changed(MODE_SELECT_BRILLOUIN, item))
        self.table_Rayleigh_regions.itemChanged.connect(
            lambda item: self.on_region_changed(MODE_SELECT_RAYLEIGH, item))

        self.setupTables()

        self.update_ui()

    def update_ui(self):
        self.refresh_plot()

    def reset_ui(self):
        self.table_Brillouin_regions.setRowCount(0)
        self.table_Rayleigh_regions.setRowCount(0)
        self.plot.cla()
        self.mplcanvas.draw()

    def on_select_brillouin_clicked(self):
        if self.mode == MODE_SELECT_BRILLOUIN:
            self.mode = MODE_DEFAULT
            self.button_brillouin_select_done.setText('Select')
            self.button_rayleigh_select_done.setText('Select')
        else:
            self.mode = MODE_SELECT_BRILLOUIN
            self.button_brillouin_select_done.setText('Done')
            self.button_rayleigh_select_done.setText('Select')

    def on_select_rayleigh_clicked(self):
        if self.mode == MODE_SELECT_RAYLEIGH:
            self.mode = MODE_DEFAULT
            self.button_brillouin_select_done.setText('Select')
            self.button_rayleigh_select_done.setText('Select')
        else:
            self.mode = MODE_SELECT_RAYLEIGH
            self.button_brillouin_select_done.setText('Select')
            self.button_rayleigh_select_done.setText('Done')

    def clear_regions(self):
        button = self.sender()
        session = Session.get_instance()
        pm = session.peak_selection_model()
        if not pm:
            return
        if button is self.button_brillouin_clear:
            pm.clear_brillouin_regions()
        elif button is self.button_rayleigh_clear:
            pm.clear_rayleigh_regions()
        self.refresh_plot()

    def on_select_data_region(self, xmin, xmax):
        if not self.plot.lines or not self.plot.lines[0]:
            return
        if self.mode == MODE_DEFAULT:
            return
        session = Session.get_instance()

        # We need the region in Hz
        region = (1e9*xmin, 1e9*xmax)

        pm = session.peak_selection_model()
        if pm:
            if self.mode == MODE_SELECT_BRILLOUIN:
                pm.add_brillouin_region(region)
            elif self.mode == MODE_SELECT_RAYLEIGH:
                pm.add_rayleigh_region(region)

        self.refresh_plot()

    def refresh_plot(self):
        self.plot.cla()
        session = Session.get_instance()
        evc = EvaluationController()

        try:
            image_key = '0'
            spectra, times, _ = evc.extract_spectra(
                image_key
            )
            if spectra is None:
                return

            cm = session.calibration_model()
            if not cm:
                return

            pm = session.peak_selection_model()
            if not pm:
                return

            if len(spectra) > 0:
                with warnings.catch_warnings():
                    warnings.filterwarnings(
                        action='ignore',
                        message='Mean of empty slice'
                    )
                    spectrum = np.nanmean(spectra, axis=0, keepdims=True)
                time = times[0]
                frequencies = cm.get_frequencies_by_time(time)
                if frequencies is not None:
                    self.plot.plot(1e-9*frequencies[0], spectrum[0])
                    self.plot.set_xlabel('$f$ [GHz]')
                    self.plot.set_xlim(1e-9*np.nanmin(frequencies),
                                       1e-9*np.nanmax(frequencies))
                self.plot.set_ylim(bottom=0)

                regions = pm.get_brillouin_regions()
                table = self.table_Brillouin_regions
                self.refresh_regions(
                    spectrum, regions, table, 'r', frequencies)

                regions = pm.get_rayleigh_regions()
                table = self.table_Rayleigh_regions
                self.refresh_regions(
                    spectrum, regions, table, 'm', frequencies)

        except Exception as e:
            logger.error('Exception occurred in peak selection: %s' % e)
        finally:
            self.mplcanvas.draw()

    def setupTables(self):
        self.table_Brillouin_regions.setColumnCount(2)
        self.table_Brillouin_regions\
            .setHorizontalHeaderLabels(["start", "end"])
        header = self.table_Brillouin_regions.horizontalHeader()
        header.setSectionResizeMode(0,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)

        self.table_Rayleigh_regions.setColumnCount(2)
        self.table_Rayleigh_regions\
            .setHorizontalHeaderLabels(["start", "end"])
        header = self.table_Rayleigh_regions.horizontalHeader()
        header.setSectionResizeMode(0,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)

    def refresh_regions(self, spectrum, regions, table, color,
                        frequencies=None):
        table.setRowCount(len(regions))
        for rowIdx, region in enumerate(regions):
            if frequencies is not None:
                ind_l = np.nanargmin(abs(frequencies[0] - region[0]))
                ind_r = np.nanargmin(abs(frequencies[0] - region[1]))
                mask = slice(ind_l, ind_r)
                self.plot.plot(
                    1e-9*frequencies[0][mask], spectrum[0][mask], color)
            # Add regions to table
            # Block signals, so the itemChanged signal is not
            # emitted during table creation
            table.blockSignals(True)
            for columnIdx, value in enumerate(region):
                item = QtWidgets.QTableWidgetItem(str(1e-9*value))
                table.setItem(rowIdx, columnIdx, item)
            table.blockSignals(False)

    def on_region_changed(self, type, item):
        row = item.row()
        column = item.column()
        value = float(item.text())

        session = Session.get_instance()

        pm = session.peak_selection_model()
        if pm:
            if type == MODE_SELECT_BRILLOUIN:
                regions = pm.get_brillouin_regions()
                current_region = np.asarray(regions[row])
                current_region[column] = 1e9*value
                current_region = tuple(current_region)
                pm.set_brillouin_region(row, current_region)
            elif type == MODE_SELECT_RAYLEIGH:
                regions = pm.get_rayleigh_regions()
                current_region = np.asarray(regions[row])
                current_region[column] = 1e9*value
                current_region = tuple(current_region)
                pm.set_rayleigh_region(row, current_region)
            self.refresh_plot()



import pathlib
from collections import namedtuple

import pytest
import numpy as np

from bmlab.session import Session

from bmicro.gui.main import BMicro


def data_file_path(file_name):
    return pathlib.Path(__file__).parent.parent / 'data' / file_name


Event = namedtuple('Event', 'xdata ydata')


@pytest.fixture
def window(mocker):
    window = BMicro()
    Session.get_instance().clear()
    file_name = data_file_path('Water.h5')

    def mock_getOpenFileName(self, *args, **kwargs):
        return file_name, None

    mocker.patch('PyQt6.QtWidgets.QFileDialog.getOpenFileName',
                 mock_getOpenFileName)
    window.open_file()
    yield window
    window.close()


def test_clicking_in_select_mode_adds_points(qtbot, window):
    ev = window.widget_extraction_view
    # Change to select mode:
    ev.toggle_mode()

    event = Event(50, 50)
    ev.on_click_image(event)
    event = Event(150, 150)
    ev.on_click_image(event)

    session = Session.get_instance()
    assert session.extraction_model().get_points('1') == [(50, 50), (150, 150)]
    assert session.extraction_model().get_points('2') == []

    # Change back to default mode
    ev.toggle_mode()

    # Clicking should not add point
    event = Event(111, 111)
    ev.on_click_image(event)
    assert session.extraction_model().get_points('1') == [(50, 50), (150, 150)]


def test_selecting_three_points_creates_circle_fit(qtbot, window):
    ev = window.widget_extraction_view
    # Change to select mode:
    ev.toggle_mode()
    event = Event(0, 100)
    ev.on_click_image(event)
    event = Event(100, 0)
    ev.on_click_image(event)
    session = Session.get_instance()
    np.testing.assert_array_equal(
        session.extraction_model().get_arc_by_calib_key('1'),
        np.empty(0)
    )
    event = Event(100/(2**0.5), 100/(2**0.5))
    ev.on_click_image(event)
    fit = session.extraction_model().get_arc_by_calib_key('1')
    assert fit is not None
    assert len(fit) == 500
    np.testing.assert_array_almost_equal(fit[0, 2, :], [0, 100])
    np.testing.assert_array_almost_equal(fit[0, 2, :], [0, 100])
    np.testing.assert_array_almost_equal(fit[-1, 2, :], [100, 0])


def test_clicking_clear_deletes_points(qtbot, window):
    ev = window.widget_extraction_view
    # Change to select mode:
    ev.toggle_mode()

    event = Event(50, 50)
    ev.on_click_image(event)

    session = Session.get_instance()
    assert session.extraction_model().get_points('1') == [(50, 50)]
    ev.clear_points()
    assert session.extraction_model().get_points('1') == []


def test_clicking_find_points_finds_points(qtbot, window):
    ev = window.widget_extraction_view
    ev.clear_points()

    session = Session.get_instance()
    assert session.extraction_model().get_points('1') == []

    ev.find_points()
    assert len(session.extraction_model().get_points('1')) > 0



def main():
    import os
    from importlib import resources
    import sys
    import logging

    from PyQt6 import QtGui, QtWidgets

    from bmicro.gui.main import BMicro
    from bmicro._version import version as __version__
    """
    Starts the BMicro application and handles its life cycle.
    """
    app = QtWidgets.QApplication(sys.argv)
    # set window icon
    ref = resources.files('bmicro') / 'img'
    with resources.as_file(ref) as imdir:
        icon_path = os.path.join(imdir, "icon.png")
        app.setWindowIcon(QtGui.QIcon(icon_path))

    window = BMicro()
    window.show()

    for arg in sys.argv:
        if arg == '--version':
            print(__version__)
            QtWidgets.QApplication.processEvents()
            sys.exit(0)
        elif arg.startswith('--log='):
            log_level = arg[6:]
            logging.basicConfig(level=log_level)

    if sys.argv[-1].endswith('.h5'):
        window.open_file(sys.argv[-1])

    sys.exit(app.exec())


if __name__ == '__main__':
    main()



import pathlib

import pytest

from bmlab.session import Session

from bmicro.gui.main import BMicro


def data_file_path(file_name):
    return pathlib.Path(__file__).parent.parent / 'data' / file_name


@pytest.fixture
def window(mocker):
    window = BMicro()
    Session.get_instance().clear()
    file_name = data_file_path('Water.h5')

    def mock_getOpenFileName(self, *args, **kwargs):
        return file_name, None

    mocker.patch('PyQt6.QtWidgets.QFileDialog.getOpenFileName',
                 mock_getOpenFileName)
    window.open_file()
    yield window
    window.close()



from importlib import resources
import logging

from PyQt6 import QtWidgets, QtCore, uic
from matplotlib.patches import Circle as MPLCircle

import matplotlib
import numpy as np

from bmlab.session import Session
from bmlab.controllers import ExtractionController

from bmicro.BGThread import BGThread
from bmicro.gui.mpl import MplCanvas


MODE_DEFAULT = 0
MODE_SELECT = 1

logger = logging.getLogger(__name__)


class ExtractionView(QtWidgets.QWidget):
    """
    Class for the extraction widget
    """

    def __init__(self, *args, **kwargs):
        super(ExtractionView, self).__init__(*args, **kwargs)

        ref = resources.files('bmicro.gui.extraction') / 'extraction_view.ui'
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self)

        self.mode = MODE_DEFAULT
        self.current_frame = 0

        self.mplcanvas = MplCanvas(
            self.image_widget, toolbar=('Home', 'Pan', 'Zoom'))
        self.image_plot = self.mplcanvas.get_figure().add_subplot(111)
        self.image_plot.axis('off')
        self.mplcanvas.get_figure().canvas.mpl_connect(
            'button_press_event', self.on_click_image)

        self.thread = BGThread()

        self.combobox_datasets.currentIndexChanged.connect(
            self.on_select_dataset)

        self.table_selected_points.itemChanged.connect(
            lambda item: self.on_points_changed(item))

        self.setupTable()

        self.button_select_done.clicked.connect(self.toggle_mode)
        self.button_clear.clicked.connect(self.clear_points)
        self.button_optimize.clicked.connect(self.optimize_points)
        self.button_find_points.clicked.connect(self.find_points)
        self.button_find_points_all.clicked.connect(self.find_points_all)

        self.button_prev_frame.clicked.connect(self.prev_frame)
        self.button_next_frame.clicked.connect(self.next_frame)

        self.update_ui()
        self.checkFrameNavigationButtons()

        self.extraction_controller = ExtractionController()

    def update_ui(self):
        session = Session.get_instance()
        self.combobox_datasets.clear()
        if not session.current_repetition():
            return

        calib_keys = session.get_calib_keys(sort_by_time=True)
        self.combobox_datasets.addItems(calib_keys)

    def reset_ui(self):
        self.combobox_datasets.clear()
        self.table_selected_points.setRowCount(0)
        self.refresh_image_plot()

    def prev_frame(self):
        if self.current_frame > 0:
            self.current_frame -= 1
            self.refresh_image_plot()
            self.checkFrameNavigationButtons()

    def next_frame(self):
        session = Session.get_instance()
        calib_key = self.combobox_datasets.currentText()
        frame_count = session.get_calibration_image_count(calib_key)
        if self.current_frame < frame_count - 1:
            self.current_frame += 1
            self.refresh_image_plot()
            self.checkFrameNavigationButtons()

    def checkFrameNavigationButtons(self):
        if self.current_frame > 0:
            self.button_prev_frame.setEnabled(True)
        else:
            self.button_prev_frame.setEnabled(False)

        session = Session.get_instance()
        calib_key = self.combobox_datasets.currentText()
        if calib_key:
            frame_count = session.get_calibration_image_count(calib_key)
            if self.current_frame < frame_count - 1:
                self.button_next_frame.setEnabled(True)
            else:
                self.button_next_frame.setEnabled(False)

    def on_select_dataset(self):
        """
        Action triggered when user selects a calibration dataset.
        """
        self.checkFrameNavigationButtons()
        self.refresh_image_plot()

    def on_click_image(self, event):
        """
        Action triggered when user clicks on preview image in selection mode.

        Parameters
        ----------
        event: matplotlib event object
            The mouse click event.
        """
        if self.mode != MODE_SELECT:
            return
        ec = ExtractionController()
        calib_key = self.combobox_datasets.currentText()
        logger.debug('Adding point (%f, %f) for calibration key %s' % (
            event.xdata, event.ydata, calib_key
        ))
        ec.add_point(calib_key,
                     (event.xdata, event.ydata))
        self.refresh_image_plot()

    def refresh_image_plot(self):
        """
        Updates the plot of the selected calibration image.
        """
        self.image_plot.cla()
        session = Session.get_instance()
        calib_key = self.combobox_datasets.currentText()
        if not calib_key:
            self.mplcanvas.draw()
            return

        em = session.extraction_model()
        img = session.get_calibration_image(calib_key, self.current_frame)

        # imshow should always get the transposed image such that
        # the horizontal axis of the plot coincides with the
        # 0-axis of the plotted array:
        self.image_plot.imshow(img.T, origin='lower', vmin=100, vmax=300)
        self.image_plot.set_title('Frame %d' % (self.current_frame+1))

        points = em.get_points(calib_key)

        if len(points) >= 3:
            em.set_image_shape(img.shape)

            arcs = em.get_arc_by_calib_key(calib_key)
            for arc in arcs:
                dr = arc[-1] - arc[0]
                line = matplotlib.patches.FancyArrow(
                    *arc[0], dr[0], dr[1], head_width=0,
                    head_length=0, color='Yellow', alpha=0.5)
                self.image_plot.add_patch(line)

        self._plot_points(em.get_points(calib_key))

        self.mplcanvas.draw()
        self.refresh_points()

    def setupTable(self):
        self.table_selected_points.setColumnCount(2)
        self.table_selected_points\
            .setHorizontalHeaderLabels(["x", "y"])
        header = self.table_selected_points.horizontalHeader()
        header.setSectionResizeMode(0,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)

    def refresh_points(self):
        session = Session.get_instance()
        em = session.extraction_model()
        calib_key = self.combobox_datasets.currentText()
        if not calib_key:
            return
        if em:
            points = em.get_points(calib_key)
            self.table_selected_points.setRowCount(len(points))
            for rowIdx, point in enumerate(points):
                # Add regions to table
                # Block signals, so the itemChanged signal is not
                # emitted during table creation
                self.table_selected_points.blockSignals(True)
                for columnIdx, value in enumerate(point):
                    item = QtWidgets.QTableWidgetItem(str(value))
                    self.table_selected_points.setItem(rowIdx, columnIdx, item)
                self.table_selected_points.blockSignals(False)

    def on_points_changed(self, item):
        row = item.row()
        column = item.column()
        value = float(item.text())

        session = Session.get_instance()
        calib_key = self.combobox_datasets.currentText()

        em = session.extraction_model()
        ec = ExtractionController()
        if em:
            points = em.get_points(calib_key)
            current_point = np.asarray(points[row])
            current_point[column] = value
            current_point = tuple(current_point)
            ec.set_point(calib_key, row, current_point)
            self.refresh_image_plot()

    def _plot_points(self, points):
        for p in points:
            circle = MPLCircle(p, radius=3, color='red', alpha=0.5)
            self.image_plot.add_patch(circle)

    def toggle_mode(self):
        """
        Toggles between normal and selection mode.

        In selection mode, the user can select points in the calibration
        image.
        """
        if self.mode == MODE_DEFAULT:
            self.mode = MODE_SELECT
            self.button_select_done.setText('Done')
        else:
            self.mode = MODE_DEFAULT
            self.button_select_done.setText('Select')

    def clear_points(self):
        """
        Deletes the selected point for the current calibration image
        """
        calib_key = self.combobox_datasets.currentText()
        session = Session.get_instance()

        if not session.extraction_model():
            return
        session.extraction_model().clear_points(calib_key)
        self.refresh_image_plot()

    def optimize_points(self):
        """
        Moves the selected points in the calibration image to the nearest
        maximum values.
        """
        calib_key = self.combobox_datasets.currentText()
        ec = ExtractionController()
        ec.optimize_points(calib_key)
        self.refresh_image_plot()

    def find_points(self):
        """
        Automatically finds the Rayleigh and Brillouin peaks of interest.
        """
        calib_key = self.combobox_datasets.currentText()
        ec = ExtractionController()
        ec.find_points(calib_key)
        self.refresh_image_plot()

    def find_points_all(self):
        """
        Automatically finds the Rayleigh and Brillouin peaks of interest
        for all calibrations existing.
        """
        session = Session.get_instance()
        calib_keys = session.get_calib_keys(sort_by_time=True)

        if not calib_keys:
            return

        for i, calib_key in enumerate(calib_keys):
            self.combobox_datasets.setCurrentText(calib_key)

            dnkw = {
                "calib_key": calib_key,
            }

            self.thread.set_task(
                func=self.extraction_controller.find_points, fkw=dnkw)
            self.thread.start()
            self.thread.wait()

            self.refresh_image_plot()
            QtCore.QCoreApplication.instance().processEvents()



from PyQt6 import QtCore


class BGThread(QtCore.QThread):
    def __init__(self, func=None, fkw=None, *args, **kwargs):
        super(BGThread, self).__init__(*args, **kwargs)
        self.func = func
        self.fkw = fkw
        self.result = None

    def set_task(self, func, fkw):
        self.func = func
        self.fkw = fkw

    def run(self):
        if self.func:
            self.result = self.func(**self.fkw)



from importlib import resources
import logging

from PyQt6 import QtWidgets, QtCore, uic
# from PyQt6.QtWidgets import QMessageBox
from matplotlib.widgets import SpanSelector
import numpy as np
import multiprocessing as mp
import time

from bmlab.session import Session

from bmlab.controllers import CalibrationController

from bmicro.BGThread import BGThread
from bmicro.gui.mpl import MplCanvas


logger = logging.getLogger(__name__)

MODE_DEFAULT = 'default'
MODE_SELECT_BRILLOUIN = 'select_brillouin_peaks'
MODE_SELECT_RAYLEIGH = 'select_rayleigh_peaks'


class CalibrationView(QtWidgets.QWidget):
    """
    Class for the calibration widget
    """

    def __init__(self, *args, **kwargs):
        super(CalibrationView, self).__init__(*args, **kwargs)

        ref = resources.files('bmicro.gui.calibration') / 'calibration_view.ui'
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self)

        self.mplcanvas = MplCanvas(self.image_widget,
                                   toolbar=('Home', 'Pan', 'Zoom'))
        self.plot = self.mplcanvas.get_figure().add_subplot(111)

        self.thread = BGThread()

        props = dict(facecolor='green', alpha=0.5)
        self.span_selector = SpanSelector(
            self.plot, onselect=self.on_select_data_region,
            useblit=True,
            direction='horizontal', props=props)

        self.button_find_peaks.clicked.connect(self.find_peaks)

        self.button_brillouin_select_done.clicked.connect(
            self.on_select_brillouin_clicked)
        self.button_rayleigh_select_done.clicked.connect(
            self.on_select_rayleigh_clicked)
        self.button_brillouin_clear.released.connect(
            self.clear_regions)
        self.button_rayleigh_clear.released.connect(
            self.clear_regions)

        self.button_fit_clear.released.connect(self.clear_calibration)

        self.button_calibrate.released.connect(self.calibrate)

        self.button_find_peaks_all.released.connect(
            lambda: self.calibrate_all(do_not='calibrate'))
        self.button_calibrate_all.released.connect(
            lambda: self.calibrate_all(do_not='find_peaks'))
        self.button_peaks_and_calibrate_all\
            .released.connect(self.calibrate_all)

        self.current_frame = 0
        self.button_prev_frame.clicked.connect(self.prev_frame)
        self.button_next_frame.clicked.connect(self.next_frame)

        self.options_dialog = None
        self.mplcanvas_options = None
        self.plot_options = None
        self.calibration_options.clicked.connect(self.show_options)

        self.mode = MODE_DEFAULT

        self.combobox_calibration.currentIndexChanged.connect(
            self.on_select_calibration)

        self.table_Brillouin_regions.itemChanged.connect(
            lambda item: self.on_region_changed(MODE_SELECT_BRILLOUIN, item))
        self.table_Rayleigh_regions.itemChanged.connect(
            lambda item: self.on_region_changed(MODE_SELECT_RAYLEIGH, item))

        self.setupTables()

        self.calibration_controller = CalibrationController()

    def update_ui(self):
        self.combobox_calibration.clear()
        session = Session.get_instance()

        if not session.file:
            return

        calib_keys = session.get_calib_keys(sort_by_time=True)
        self.combobox_calibration.addItems(calib_keys)
        self.refresh_plot()

    def reset_ui(self):
        self.table_Brillouin_regions.setRowCount(0)
        self.table_Rayleigh_regions.setRowCount(0)
        self.combobox_calibration.clear()
        self.calibration_progress.setValue(0)
        self.plot.cla()
        self.mplcanvas.draw()

    def prev_frame(self):
        if self.current_frame > 0:
            self.current_frame -= 1
            self.refresh_plot()
            self.checkFrameNavigationButtons()

    def next_frame(self):
        session = Session.get_instance()
        calib_key = self.combobox_calibration.currentText()
        frame_count = session.get_calibration_image_count(calib_key)
        if self.current_frame < frame_count - 1:
            self.current_frame += 1
            self.refresh_plot()
            self.checkFrameNavigationButtons()

    def checkFrameNavigationButtons(self):
        if self.current_frame > 0:
            self.button_prev_frame.setEnabled(True)
        else:
            self.button_prev_frame.setEnabled(False)

        session = Session.get_instance()
        calib_key = self.combobox_calibration.currentText()
        if calib_key:
            frame_count = session.get_calibration_image_count(calib_key)
            if self.current_frame < frame_count - 1:
                self.button_next_frame.setEnabled(True)
            else:
                self.button_next_frame.setEnabled(False)

    def clear_calibration(self):
        cc = CalibrationController()
        calib_key = self.combobox_calibration.currentText()
        cc.clear_calibration(calib_key)
        self.refresh_plot()

    def on_select_brillouin_clicked(self):
        if self.mode == MODE_SELECT_BRILLOUIN:
            self.mode = MODE_DEFAULT
            self.button_brillouin_select_done.setText('Select')
            self.button_rayleigh_select_done.setText('Select')
        else:
            self.mode = MODE_SELECT_BRILLOUIN
            self.button_brillouin_select_done.setText('Done')
            self.button_rayleigh_select_done.setText('Select')

    def on_select_rayleigh_clicked(self):
        if self.mode == MODE_SELECT_RAYLEIGH:
            self.mode = MODE_DEFAULT
            self.button_brillouin_select_done.setText('Select')
            self.button_rayleigh_select_done.setText('Select')
        else:
            self.mode = MODE_SELECT_RAYLEIGH
            self.button_brillouin_select_done.setText('Select')
            self.button_rayleigh_select_done.setText('Done')

    def clear_regions(self):
        button = self.sender()
        session = Session.get_instance()
        cm = session.calibration_model()
        if not cm:
            return
        calib_key = self.combobox_calibration.currentText()
        if button is self.button_brillouin_clear:
            cm.clear_brillouin_regions(calib_key)
        elif button is self.button_rayleigh_clear:
            cm.clear_rayleigh_regions(calib_key)
        self.refresh_plot()

    def on_select_data_region(self, xmin, xmax):
        if not self.plot.lines or not self.plot.lines[0]:
            return
        # Since we might operate on a frequency axis,
        # we need the indices instead of the values.
        xdata = self.plot.lines[0].get_xdata()
        indmin, indmax = np.searchsorted(xdata, (xmin, xmax))

        if self.mode == MODE_DEFAULT:
            return
        session = Session.get_instance()
        calib_key = self.combobox_calibration.currentText()

        cm = session.calibration_model()
        if cm:
            if self.mode == MODE_SELECT_BRILLOUIN:
                cm.add_brillouin_region(calib_key, (indmin, indmax))
            elif self.mode == MODE_SELECT_RAYLEIGH:
                cm.add_rayleigh_region(calib_key, (indmin, indmax))

        self.refresh_plot()

    def on_select_calibration(self):
        self.checkFrameNavigationButtons()
        self.refresh_plot()

    def calibrate(self):
        calib_key = self.combobox_calibration.currentText()

        count = mp.Value('I', 0, lock=True)
        max_count = mp.Value('i', 0, lock=True)

        dnkw = {
            "calib_key": calib_key,
            "count": count,
            "max_count": max_count,
        }

        self.thread.set_task(
            func=self.calibration_controller.calibrate, fkw=dnkw)
        self.thread.start()
        # Show a progress until computation is done
        while max_count.value == 0 or count.value < max_count.value:
            time.sleep(.05)
            self.calibration_progress.setValue(count.value)
            if max_count.value >= 0:
                self.calibration_progress.setMaximum(max_count.value)
            QtCore.QCoreApplication.instance().processEvents()
        # make sure the thread finishes
        self.thread.wait()

        self.refresh_plot()

    def calibrate_all(self, do_not=None):
        session = Session.get_instance()
        calib_keys = session.get_calib_keys(sort_by_time=True)

        if not calib_keys:
            return

        self.calibration_progress.setMaximum(len(calib_keys))

        for i, calib_key in enumerate(calib_keys):
            self.combobox_calibration.setCurrentText(calib_key)

            dnkw = {
                "calib_key": calib_key,
            }

            if do_not != 'find_peaks':
                self.thread.set_task(
                    func=self.calibration_controller.find_peaks, fkw=dnkw)
                self.thread.start()
                self.thread.wait()
            if do_not != 'calibrate':
                self.thread.set_task(
                    func=self.calibration_controller.calibrate, fkw=dnkw)
                self.thread.start()
                self.thread.wait()
            self.calibration_progress.setValue(i + 1)

            self.refresh_plot()
            QtCore.QCoreApplication.instance().processEvents()

    def refresh_plot(self):
        self.plot.cla()
        session = Session.get_instance()
        calib_key = self.combobox_calibration.currentText()
        if not calib_key:
            return

        try:
            cc = CalibrationController()
            spectrum, _, _ = cc.extract_spectra(
                calib_key,
                frame_num=self.current_frame
            )
            if spectrum is None:
                return

            cm = session.calibration_model()
            if not cm:
                return

            if len(spectrum) > 0:
                spectrum = spectrum[0]
                frequencies = cm.get_frequencies_by_calib_key(calib_key)
                frequency = None
                if frequencies:
                    frequency = 1e-9*frequencies[self.current_frame]
                    self.plot.plot(frequency, spectrum)
                    self.plot.set_xlabel('$f$ [GHz]')
                    self.plot.set_xlim(1e-9*np.min(frequencies),
                                       1e-9*np.max(frequencies))
                else:
                    self.plot.plot(spectrum)
                    self.plot.set_xlabel('$f$ [pix]')
                    self.plot.set_xlim(0, len(spectrum))
                self.plot.set_ylim(bottom=0)
                self.plot.set_title('Frame %d' %
                                    (self.current_frame+1))

                regions = cm.get_brillouin_regions(calib_key)
                table = self.table_Brillouin_regions
                self.refresh_regions(spectrum, regions, table, 'r', frequency)

                for region_key, region in enumerate(regions):
                    fit = cm.brillouin_fits\
                        .get_fit(calib_key, region_key, self.current_frame)
                    if fit is not None:
                        w0s = fit.w0s
                        w0s_f = cm.get_frequency_by_calib_key(w0s, calib_key)
                        if w0s_f is not None:
                            self.plot.vlines(1e-9*w0s_f[0], 0, np.nanmax(
                                spectrum), colors=['black'])
                            self.plot.vlines(1e-9*w0s_f[1], 0, np.nanmax(
                                spectrum), colors=['black'])
                        else:
                            self.plot.vlines(w0s[0], 0, np.nanmax(
                                spectrum), colors=['black'])
                            self.plot.vlines(w0s[1], 0, np.nanmax(
                                spectrum), colors=['black'])

                regions = cm.get_rayleigh_regions(calib_key)
                table = self.table_Rayleigh_regions
                self.refresh_regions(spectrum, regions, table, 'm', frequency)

                for region_key, region in enumerate(regions):
                    fit = cm.rayleigh_fits\
                        .get_fit(calib_key, region_key, self.current_frame)
                    if fit is not None:
                        w0 = fit.w0
                        w0_f = cm.get_frequency_by_calib_key(w0, calib_key)
                        if w0_f is not None:
                            self.plot.vlines(1e-9*w0_f, 0, np.nanmax(
                                spectrum), colors=['black'])
                        else:
                            self.plot.vlines(w0, 0, np.nanmax(
                                spectrum), colors=['black'])

                expected = cc.expected_frequencies(
                    calib_key, self.current_frame)
                if expected is not None:
                    self.plot.vlines(1e-9 * expected, 0, np.nanmax(
                        spectrum), colors=['green'])

        except Exception as e:
            logger.error('Exception occurred in calibration: %s' % e)
        finally:
            self.mplcanvas.draw()

    def setupTables(self):
        self.table_Brillouin_regions.setColumnCount(2)
        self.table_Brillouin_regions\
            .setHorizontalHeaderLabels(["start", "end"])
        header = self.table_Brillouin_regions.horizontalHeader()
        header.setSectionResizeMode(0,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)

        self.table_Rayleigh_regions.setColumnCount(2)
        self.table_Rayleigh_regions\
            .setHorizontalHeaderLabels(["start", "end"])
        header = self.table_Rayleigh_regions.horizontalHeader()
        header.setSectionResizeMode(0,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)
        header.setSectionResizeMode(1,
                                    QtWidgets.QHeaderView.ResizeMode.Stretch)

    def refresh_regions(self, spectrum, regions, table, color,
                        frequencies=None):
        table.setRowCount(len(regions))
        for rowIdx, region in enumerate(regions):
            mask = np.arange(int(region[0]), int(region[1]))
            if frequencies is not None:
                self.plot.plot(frequencies[mask], spectrum[mask], color)
            else:
                self.plot.plot(mask, spectrum[mask], color)
            # Add regions to table
            # Block signals, so the itemChanged signal is not
            # emitted during table creation
            table.blockSignals(True)
            for columnIdx, value in enumerate(region):
                item = QtWidgets.QTableWidgetItem(str(value))
                table.setItem(rowIdx, columnIdx, item)
            table.blockSignals(False)

    def on_region_changed(self, type, item):
        row = item.row()
        column = item.column()
        value = float(item.text())

        session = Session.get_instance()
        calib_key = self.combobox_calibration.currentText()

        cm = session.calibration_model()
        if cm:
            if type == MODE_SELECT_BRILLOUIN:
                regions = cm.get_brillouin_regions(calib_key)
                current_region = np.asarray(regions[row])
                current_region[column] = value
                current_region = tuple(current_region)
                cm.set_brillouin_region(calib_key, row, current_region)
            elif type == MODE_SELECT_RAYLEIGH:
                regions = cm.get_rayleigh_regions(calib_key)
                current_region = np.asarray(regions[row])
                current_region[column] = value
                current_region = tuple(current_region)
                cm.set_rayleigh_region(calib_key, row, current_region)
            self.refresh_plot()

    def find_peaks(self):
        """
        Automatically finds the Rayleigh and Brillouin peaks.
        """
        calib_key = self.combobox_calibration.currentText()
        cc = CalibrationController()
        cc.find_peaks(calib_key)
        self.refresh_plot()

    def show_options(self):
        self.options_dialog = QtWidgets.QDialog(
            self,
            QtCore.Qt.WindowType.WindowTitleHint |
            QtCore.Qt.WindowType.WindowCloseButtonHint
        )
        ref = (resources.files('bmicro.gui.calibration') /
               'calibration_options.ui')
        with resources.as_file(ref) as ui_file:
            uic.loadUi(ui_file, self.options_dialog)
        self.options_dialog.setWindowTitle('Calibration options')
        self.options_dialog.setWindowModality(
            QtCore.Qt.WindowModality.ApplicationModal)
        self.options_dialog.button_ok.clicked.connect(
            self.apply_and_close_options
        )
        self.options_dialog.button_apply.clicked.connect(
            self.apply_options
        )
        self.options_dialog.button_cancel.clicked.connect(
            self.close_options
        )
        self.options_dialog.temperature.valueChanged.connect(
            self.temperature_changed
        )
        self.options_dialog.adjustSize()

        self.mplcanvas_options = MplCanvas(self.options_dialog.widget_plot,
                                           toolbar=('Home', 'Pan', 'Zoom'))
        self.plot_options =\
            self.mplcanvas_options.get_figure().add_subplot(111)

        self.update_options_view()
        self.options_dialog.exec()

    def apply_and_close_options(self):
        self.apply_options()
        self.close_options()

    def apply_options(self):
        session = Session.get_instance()
        cm = session.calibration_model()
        if cm is None:
            return

        shift_0_old = session.setup.calibration.shift_methanol
        shift_1_old = session.setup.calibration.shift_water
        shift_0_new = 1e9 * self.options_dialog.shift_0.value()
        shift_1_new = 1e9 * self.options_dialog.shift_1.value()

        if (shift_0_old != shift_0_new) | (shift_1_old != shift_1_new):
            # Set current calibration values
            session.setup.calibration.set_shift_methanol(shift_0_new)
            session.setup.calibration.set_shift_water(shift_1_new)

            # Apply new calibration values
            calib_keys = session.get_calib_keys()
            for calib_key in calib_keys:
                self.calibration_controller.calibrate(calib_key)

            self.update_options_view()
            self.refresh_plot()

    def close_options(self):
        self.options_dialog.close()

    def temperature_changed(self):
        temperature = self.sender().value()

        session = Session.get_instance()
        # Don't do anything if the same temperature is set
        if temperature == session.setup.temperature - 273.15:
            return
        session.setup.set_temperature(temperature)
        self.update_options_view()

    def update_options_view(self):
        if self.options_dialog is None:
            return

        session = Session.get_instance()
        cm = session.calibration_model()
        if cm is None:
            return

        # Set current calibration values
        self.options_dialog.shift_0.setValue(
            1e-9 * session.setup.calibration.shift_methanol)
        self.options_dialog.shift_1.setValue(
            1e-9 * session.setup.calibration.shift_water)
        self.options_dialog.temperature.blockSignals(True)
        self.options_dialog.temperature.setValue(
            session.setup.temperature - 273.15
        )
        self.options_dialog.temperature.blockSignals(False)

        # Get the sorted calibration keys
        calib_keys = session.get_calib_keys(sort_by_time=True)
        if calib_keys is None:
            return

        # Allocate the shifts array
        shifts = np.empty((
            len(calib_keys)
            * session.get_calibration_image_count(calib_keys[0]),
            session.setup.calibration.num_brillouin_samples * 2
        ))
        shifts[:] = np.nan

        # Get the resulting fitted calibration frequencies
        cal_image = 0
        for calib_key in calib_keys:
            frame_count = session.get_calibration_image_count(calib_key)
            for frame in range(frame_count):
                cal_image = cal_image + 1
                w0s = cm.get_sorted_peaks(calib_key, frame)

                if w0s is None or w0s.shape[0]\
                        != session.setup.calibration.\
                        num_brillouin_samples * 2 + 2:
                    continue

                w0s_f = cm.get_frequency_by_calib_key(w0s, calib_key)

                if w0s_f is None:
                    continue

                for i in range(
                        session.setup.calibration.num_brillouin_samples):
                    shifts[cal_image - 1, i] =\
                        1e-9 * abs(w0s_f[0] - w0s_f[i + 1])
                    shifts[cal_image - 1, -1*i - 1] =\
                        1e-9 * abs(w0s_f[-1] - w0s_f[-1*i - 2])

        # Clear the calibration plot
        self.plot_options.cla()
        # Plot calibration reference frequency shifts
        self.plot_options.hlines(
            1e-9 * session.setup.calibration.shift_methanol,
            0, len(shifts) - 1, colors=['black'])
        if session.setup.calibration.num_brillouin_samples > 1:
            self.plot_options.hlines(
                1e-9 * session.setup.calibration.shift_water,
                0, len(shifts) - 1, colors=['black'])

        # Plot the fitted calibration frequency shifts
        self.plot_options.plot(shifts)

        # Set the axis labels
        self.plot_options.set_ylabel('$\\nu_\\mathrm{B}$ [GHz]')
        # Update the plot
        self.mplcanvas_options.draw()



import pathlib
import os

from PyQt6.QtWidgets import QWidget
from PyQt6 import QtCore

from bmlab.session import Session

from bmicro.gui.main import BMicro, check_event_mime_data


def data_file_path(file_name):
    return pathlib.Path(__file__).parent.parent / 'data' / file_name


def test_main_window_can_activate_all_tabs(qtbot):

    window = BMicro()
    qtbot.add_widget(window)
    tab_texts = ['Data', 'Extraction', 'Calibration',
                 'Peak Selection', 'Evaluation']
    tab_names = ['tab_data', 'tab_extraction', 'tab_calibration',
                 'tab_peak_selection', 'tab_evaluation']

    for idx, (tab_text, tab_name) in enumerate(zip(tab_texts, tab_names)):
        current_tab = window.tabWidget.findChild(QWidget, tab_name)
        assert current_tab
        qtbot.mouseClick(current_tab, QtCore.Qt.MouseButton.LeftButton)
        assert window.tabWidget.tabText(idx) == tab_text

    window.close()


def test_open_file_shows_metadata(qtbot, mocker):
    window = BMicro()
    file_name = 'Water.h5'
    file_path = data_file_path(file_name)

    def mock_getOpenFileName(self, *args, **kwargs):
        return file_path, None

    mocker.patch('PyQt6.QtWidgets.QFileDialog.getOpenFileName',
                 mock_getOpenFileName)

    window.open_file()
    w = window.widget_data_view
    assert w.label_selected_file.text() == str(file_name)
    assert w.label_selected_file.toolTip() == str(file_path)
    assert w.label_date.text() == '2020-11-03 15:20'
    assert w.label_resolution_x.text() == '10'
    assert w.label_resolution_y.text() == '1'
    assert w.label_resolution_z.text() == '1'
    assert w.label_calibration.text() == 'True'
    assert w.textedit_comment.toPlainText() == 'Brillouin data'

    window.close()


def test_check_event_mime_data():
    class DummyEvent:
        def __init__(self, *args, **kwargs):
            self.__mimeData = QtCore.QMimeData()

        def mimeData(self):
            return self.__mimeData

    event = DummyEvent()

    """ Test for empty url """
    path = check_event_mime_data(event)
    assert path is False

    """ Test for wrong file type """
    event.mimeData().setUrls([QtCore.QUrl("file:/directory/file.txt")])
    path = check_event_mime_data(event)
    assert path is False

    """ Test for correct file type """
    event.mimeData().setUrls([QtCore.QUrl("file:/directory/file.h5")])
    path = check_event_mime_data(event)
    assert path == '/directory/file.h5'


def test_save_and_load_session():
    if os.path.exists(data_file_path('Water.session.h5')):
        os.remove(data_file_path('Water.session.h5'))
    session = Session.get_instance()
    session.set_file(data_file_path('Water.h5'))
    session.set_reflection(vertically=True)
    session.extraction_models['0'].add_point('0', 10, 30, 30)
    session.save()

    assert os.path.exists(data_file_path('Water.session.h5'))

    session.clear()

    assert session.file is None
    assert session.extraction_models == {}

    session.set_file(data_file_path('Water.h5'))

    assert session.file is not None
    assert len(session.extraction_models) == 1

    os.remove(data_file_path('Water.session.h5'))
